# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
from hanabython.Modules.Action import Action
from hanabython.Modules.StringAnsi import StringAnsi


class ActionForfeit(Action):
    """
    An action of a player: forfeit (lose the game immediately).

    >>> action = ActionForfeit()
    >>> print(action)
    Forfeit
    """

    def __init__(self):
        super().__init__(Action.FORFEIT)

    def colored(self) -> str:
        return (StringAnsi.RED + StringAnsi.STYLE_BOLD
                + 'Forfeit' + StringAnsi.RESET)


if __name__ == '__main__':
    my_action = ActionForfeit()
    my_action.test_str()

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
from hanabython.Modules.Clue import Clue
from hanabython.Modules.Colored import Colored
from hanabython.Modules.Color import Color
from hanabython.Modules.Colors import Colors


class Card(Colored):
    """
    A card of Hanabi.

    :param Color c: the color of the card.
    :param int v: the value of the card (usually between 1 and 5).
    :param str s: a short string representing the card. Must use one of the
        standard colors, cf. :meth:`Color.from_symbol`.

    You can provide either :attr:`c` and :attr:`v`, or :attr:`s`.
    The constructor accepts several types of syntax, as illustrated below.

    >>> my_card = Card(c=Colors.BLUE, v=3)
    >>> print(my_card)
    B3
    >>> my_card = Card(Colors.BLUE, 3)
    >>> print(my_card)
    B3
    >>> my_card = Card(3, Colors.BLUE)
    >>> print(my_card)
    B3
    >>> my_card = Card(s='B3')
    >>> print(my_card)
    B3
    >>> my_card = Card('B3')
    >>> print(my_card)
    B3
    >>> my_card = Card(s='3B')
    >>> print(my_card)
    B3
    >>> my_card = Card('3B')
    >>> print(my_card)
    B3

    N.B.: the string input works even if the v has several digits.

    >>> my_card = Card('B42')
    >>> print(my_card)
    B42
    >>> my_card = Card('51M')
    >>> print(my_card)
    M51
    """
    def __init__(self, *args, **kwargs):
        s = None
        if len(args) == 1:
            s = args[0]
        elif len(args) == 2:
            if type(args[0]) == int:
                self.v = args[0]
                self.c = args[1]
            elif type(args[1]) == int:
                self.c = args[0]
                self.v = args[1]
            else:
                raise ValueError('One argument should be an integer.')
        elif 's' in kwargs.keys():
            s = kwargs['s']
        else:
            self.c = kwargs['c']
            self.v = kwargs['v']
        if s is not None:
            try:
                self.v = int(s[1:])
                self.c = Colors.from_symbol(s[0])
            except ValueError:
                try:
                    self.v = int(s[:-1])
                    self.c = Colors.from_symbol(s[-1])
                except ValueError:
                    raise ValueError('Could not interpret as a card: ', s)

    def colored(self) -> str:
        return self.c.color_str(self.c.symbol + str(self.v))

    def match(self, clue: Clue) -> bool:
        """
        React to a clue.

        :param clue: the clue.

        :return: whether the card should be pointed when giving this clue.

        >>> from hanabython.Modules.Colors import Colors
        >>> card_blue = Card('B3')
        >>> card_blue.match(Clue(Colors.BLUE))
        True
        >>> card_blue.match(Clue(Colors.RED))
        False
        >>> card_blue.match(Clue(3))
        True
        >>> card_blue.match(Clue(4))
        False
        >>> card_multi = Card('M3')
        >>> card_multi.match(Clue(Colors.BLUE))
        True
        >>> card_colorless = Card('C3')
        >>> card_colorless.match(Clue(Colors.BLUE))
        False
        """
        if clue.category == Clue.VALUE:
            return self.v == clue.x
        else:
            return self.c.match(clue.x)


if __name__ == '__main__':
    card = Card(c=Colors.BLUE, v=3)
    card.test_str()

    print('\nIs is blue?', card.match(Clue(Colors.BLUE)))
    print('Is it a 4?', card.match(Clue(4)))

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
from hanabython.Modules.Colored import Colored
from hanabython.Modules.StringAnsi import StringAnsi


class Color(Colored):
    r"""
    A color in Hanabi.

    :param name: The full name of the color. In a game, two distinct
        colors must have different names.
    :param symbol: The short name of the color. For standard colors
        (defined as constants in :class:`Colors`), it is always 1 character,
        and no two standard colors have the same symbol. For user-defined
        colors, it is recommended to do the same, but not necessary.
    :param print_color: an ANSI escape code that modifies the printing
        color. See :class:`StringAnsi`.

    >>> brown = Color(name='Brown', symbol='N', print_color=StringAnsi.BROWN)
    >>> brown.name
    'Brown'
    >>> brown.symbol
    'N'
    >>> brown.print_color
    '\x1b[33m'
    """

    def __init__(self, name: str, symbol: str, print_color: str):
        self.name = name
        self.symbol = symbol
        self.print_color = print_color

    def colored(self) -> str:
        return self.color_str(self.symbol)

    def color_str(self, o: object) -> str:
        r"""
        Convert an object to a colored string.

        :param o: any object.

        :return: the ``__str__`` of this object, with an ANSI color-modifying
            escape code at the beginning and its cancellation at the end.

        >>> brown = Color(name='Brown', symbol='N',
        ...               print_color=StringAnsi.BROWN)
        >>> brown.color_str('some text')
        '\x1b[33msome text\x1b[0;0m'
        >>> brown.color_str(42)
        '\x1b[33m42\x1b[0;0m'
        """
        return self.print_color + str(o) + StringAnsi.RESET

    def __eq__(self, other):
        return isinstance(other, Color) and (
            (self.name, self.symbol, self.print_color)
            == (other.name, other.symbol, other.print_color)
        )

    def __hash__(self):
        return hash((self.name, self.symbol, self.print_color))

    def match(self, clue_color: 'Color') -> bool:
        """
        React to a color clue.

        :param clue_color: the color of the clue.

        :return: whether a card of the current color should react to a clue of
            color :attr:`clue_color`. A normal color matches simply if the color
            of the clue is the same. This is different in
            :class:`ColorMulticolor` and :class:`ColorColorless`.

        >>> brown = Color(name='Brown', symbol='N',
        ...               print_color=StringAnsi.BROWN)
        >>> pink = Color(name='Pink', symbol='P',
        ...              print_color=StringAnsi.MAGENTA)
        >>> brown.match(clue_color=brown)
        True
        >>> brown.match(clue_color=pink)
        False
        """
        return self == clue_color

    @property
    def is_cluable(self):
        """
        :return: whether this color can be used for clues. For a normal color,
            it is True. This is different in
            :class:`ColorMulticolor` and :class:`ColorColorless`.
        """
        return True


if __name__ == '__main__':
    brown = Color(name='Brown', symbol='N', print_color=StringAnsi.BROWN)
    brown.test_str()

    pink = Color(name='Pink', symbol='P', print_color=StringAnsi.MAGENTA)
    pseudo_pink = Color(name='Pink', symbol='P', print_color=StringAnsi.MAGENTA)

    print('pseudo_pink == pink: ', pseudo_pink == pink)
    my_set = {brown, pink}
    print('In the set = ', pseudo_pink in my_set)
    my_dict = {brown: 'a', pink: 'b'}
    print('In the dict = ', pseudo_pink in my_dict.keys())
    print('my_dict[pseudo_pink] = ', my_dict[pseudo_pink])

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
from hanabython.Modules.Colored import Colored
from typing import Iterable


class ConfigurationColorContents(Colored, list):
    """
    The contents of a color in a deck of Hanabi.

    This is essentially a list, stating the number of copies for each card.
    For example, [3, 2, 2, 2, 1] means there are 3 ones, 2 twos, etc.
    Each integer in this list must be strictly positive.

    :param contents: an iterable used to create the list.
    :param name: the name of the configuration. Can be None (default value).
        Should not be capitalized (e.g. "my favorite configuration" and not
        "My favorite configuration").

    >>> cfg = ConfigurationColorContents.NORMAL
    >>> print(cfg.name)
    normal
    >>> print(cfg)
    normal
    >>> print(list(cfg))
    [3, 2, 2, 2, 1]
    >>> cfg = ConfigurationColorContents([3, 2, 1])
    >>> print(cfg.name)
    None
    >>> print(cfg)
    [3, 2, 1]
    """

    def __init__(self, contents: Iterable[int], name: str = None):
        super(ConfigurationColorContents, self).__init__(contents)
        self.name = name

    def colored(self) -> str:
        if self.name is None:
            return str(list(self))
        else:
            return self.name

    #: Normal contents of a color (3 2 2 2 1).
    NORMAL = None
    #: Contents of a short color (1 1 1 1 1).
    SHORT = None


ConfigurationColorContents.NORMAL = ConfigurationColorContents(
    contents=[3, 2, 2, 2, 1], name='normal'
)
ConfigurationColorContents.SHORT = ConfigurationColorContents(
    contents=[1, 1, 1, 1, 1], name='short'
)


if __name__ == '__main__':
    my_cfg = ConfigurationColorContents.NORMAL
    my_cfg.test_str()
    print(list(my_cfg))

    print('\n')
    my_cfg = ConfigurationColorContents(
        contents=[3, 2, 1]
    )
    my_cfg.test_str()

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
from typing import List
import numpy as np
from hanabython.Modules.Colored import Colored
from hanabython.Modules.StringUtils import uncolor, str_from_iterable
from hanabython.Modules.Configuration import Configuration
from hanabython.Modules.Card import Card


class DiscardPile(Colored):
    """
    The discard pile in a game of Hanabi.

    :param cfg: the configuration of the game.

    :var list chronological: a list a cards discarded, by chronological order.
    :var np.array array: each row represents a color, each column a card value.
        The coefficient is the number of copies of this card in the discard
        pile.
    :var np.array not_discarded: is equal to :attr:`Configuration.deck_array`
        - :attr:`array`. Number of copies left for each card (including
        everything except the discard pile: the draw pile, the players' hand and
        the board).
    :var np.array scorable: each row represents a color, each column a
        card value. The coefficient is True it is possible to have a such
        card on the board at the end of the game (whether it is already on the
        board or not). For example, if the two G4's are discarded, then G4 and
        G5 are not "scorable". Note that a 1 always is considered "scorable",
        whether it is on the board or not.

    >>> from hanabython.Modules.Configuration import Configuration
    >>> discard_pile = DiscardPile(Configuration.STANDARD)
    >>> print(discard_pile)
    No card discarded yet

    Check that scorable cards are counted correctly with unusual configurations:

    >>> from hanabython.Modules.Configuration import Configuration
    >>> from hanabython.Modules.ConfigurationDeck import ConfigurationDeck
    >>> from hanabython.Modules.Colors import Colors
    >>> from hanabython.Modules.ConfigurationColorContents import (
    ...     ConfigurationColorContents)
    >>> discard_pile = DiscardPile(Configuration(
    ...     deck=ConfigurationDeck(contents=[
    ...         (Colors.BLUE, ConfigurationColorContents([3, 2, 1, 1])),
    ...         (Colors.RED, ConfigurationColorContents([2, 1])),
    ...     ])
    ... ))
    >>> print(discard_pile)
    No card discarded yet
    >>> print(discard_pile.array)
    [[0 0 0 0]
     [0 0 0 0]]
    >>> print(discard_pile.not_discarded)
    [[3 2 1 1]
     [2 1 0 0]]
    >>> print(discard_pile.scorable)
    [[ True  True  True  True]
     [ True  True False False]]
    >>> print(discard_pile.max_score_possible)
    6
    """

    def __init__(self, cfg: Configuration):
        self.cfg = cfg
        self.chronological = []
        self.array = np.zeros(cfg.deck_array.shape, dtype=int)
        self.not_discarded = np.copy(cfg.deck_array)
        # This formula below is valid only at the beginning because there is no
        # "holes" (zeros) in the middle of some rows.
        self.scorable = (cfg.deck_array > 0)

    def __repr__(self) -> str:
        return '<DiscardPile: %s>' % self.str_compact_chronological()

    def colored(self) -> str:
        return self.colored_multi_line_compact()

    @property
    def max_score_possible(self):
        """
        Maximum possible score, considering the discard pile.

        :return: the maximum score that is still possible.
        """
        return np.sum(self.scorable)

    def str_multi_line_compact(self) -> str:
        """
        Convert to nice string.

        :return: a representation of the discard pile. As of now, it is the
            one used for the standard method :meth:`__str__` (this behavior
            might be modified in the future).

        >>> from hanabython.Modules.Configuration import Configuration
        >>> discard_pile = DiscardPile(Configuration.STANDARD)
        >>> discard_pile.receive(Card('B3'))
        >>> discard_pile.receive(Card('R4'))
        >>> discard_pile.receive(Card('B1'))
        >>> print(discard_pile.str_multi_line_compact())
        B1 B3
        R4
        """
        return uncolor(self.colored_multi_line_compact())

    def colored_multi_line_compact(self) -> str:
        """
        Colored version of :meth:`str_multi_line_compact`.
        """
        if len(self.chronological) == 0:
            return 'No card discarded yet'
        lines = []
        for i, c in enumerate(self.cfg.colors):
            if np.sum(self.array[i, :]) == 0:
                continue
            words = [str(Card(c, v))
                     for j, v in enumerate(self.cfg.values)
                     for _ in range(self.array[i, j])]
            lines.append(c.color_str(' '.join(words)))
        return '\n'.join(lines)

    def str_multi_line(self) -> str:
        """
        Convert to nice string.

        :return: a representation of the discard pile.

        >>> from hanabython.Modules.Configuration import Configuration
        >>> discard_pile = DiscardPile(Configuration.STANDARD)
        >>> discard_pile.receive(Card('B3'))
        >>> discard_pile.receive(Card('R4'))
        >>> discard_pile.receive(Card('B1'))
        >>> print(discard_pile.str_multi_line())
        B1 B3
        -
        R4
        -
        -
        """
        return uncolor(self.colored_multi_line())

    def colored_multi_line(self) -> str:
        """
        Colored version of :meth:`str_multi_line`.
        """
        lines = []
        for i, c in enumerate(self.cfg.colors):
            if np.sum(self.array[i, :]) == 0:
                lines.append(c.color_str('-'))
                continue
            words = [str(Card(c, v))
                     for j, v in enumerate(self.cfg.values)
                     for _ in range(self.array[i, j])]
            lines.append(c.color_str(' '.join(words)))
        return '\n'.join(lines)

    def str_as_array(self) -> str:
        """
        Convert to string in an array-style layout.

        :return: a representation of the discard pile.

        >>> from hanabython.Modules.Configuration import Configuration
        >>> discard_pile = DiscardPile(Configuration.STANDARD)
        >>> discard_pile.receive(Card('B3'))
        >>> discard_pile.receive(Card('R4'))
        >>> discard_pile.receive(Card('B1'))
        >>> print(discard_pile.str_as_array())
           1 2 3 4 5
        B [1 0 1 0 0]
        G [0 0 0 0 0]
        R [0 0 0 1 0]
        W [0 0 0 0 0]
        Y [0 0 0 0 0]
        """
        return uncolor(self.colored_as_array())

    def colored_as_array(self) -> str:
        """
        Colored version of :meth:`str_as_array`.
        """
        to_join = [
            '   ' + ' '.join([str(i + 1) for i in range(self.cfg.n_values)])
        ]
        for i, c in enumerate(self.cfg.colors):
            to_join.append(
                c.color_str('%s %s' % (c.symbol, self.array[i, :]))
            )
        return '\n'.join(to_join)

    def str_compact_factorized(self) -> str:
        """
        Convert to nice string.

        :return: a representation of the discard pile.

        >>> from hanabython.Modules.Configuration import Configuration
        >>> discard_pile = DiscardPile(Configuration.STANDARD)
        >>> discard_pile.receive(Card('B3'))
        >>> discard_pile.receive(Card('R4'))
        >>> discard_pile.receive(Card('B1'))
        >>> print(discard_pile.str_compact_factorized())
        B 1 3  R 4
        """
        return uncolor(self.colored_compact_factorized())

    def colored_compact_factorized(self) -> str:
        """
        Colored version of :meth:`str_multi_line_compact`.
        """
        if len(self.chronological) == 0:
            return 'No card discarded yet'
        lines = []
        for i, c in enumerate(self.cfg.colors):
            if np.sum(self.array[i, :]) == 0:
                continue
            words = [str(v)
                     for j, v in enumerate(self.cfg.values)
                     for _ in range(self.array[i, j])]
            lines.append(c.color_str(c.symbol + ' ' + ' '.join(words)))
        return '  '.join(lines)

    def str_compact_ordered(self) -> str:
        """
        Convert to string in a list-style layout, ordered by color and value.

        :return: a representation of the discard pile.

        >>> from hanabython.Modules.Configuration import Configuration
        >>> discard_pile = DiscardPile(Configuration.STANDARD)
        >>> discard_pile.receive(Card('B3'))
        >>> discard_pile.receive(Card('R4'))
        >>> discard_pile.receive(Card('B1'))
        >>> print(discard_pile.str_compact_ordered())
        B1 B3 R4
        """
        return uncolor(self.colored_compact_ordered())

    def colored_compact_ordered(self) -> str:
        """
        Colored version of :meth:`str_compact_ordered`.
        """
        if not self.chronological:
            return 'No card discarded yet'
        return str_from_iterable(
            [card.colored() for card in self.list_reordered])

    def str_compact_chronological(self) -> str:
        """
        Convert to string in a list-style layout, by chronological order.

        :return: a representation of the discard pile.

        >>> from hanabython.Modules.Configuration import Configuration
        >>> discard_pile = DiscardPile(Configuration.STANDARD)
        >>> discard_pile.receive(Card('B3'))
        >>> discard_pile.receive(Card('R4'))
        >>> discard_pile.receive(Card('B1'))
        >>> print(discard_pile.str_compact_chronological())
        B3 R4 B1
        """
        return uncolor(self.colored_compact_chronological())

    def colored_compact_chronological(self) -> str:
        """
        Colored version of :meth:`str_compact_chronological`.
        """
        if not self.chronological:
            return 'No card discarded yet'
        return str_from_iterable(
            [card.colored() for card in self.chronological])

    @property
    def list_reordered(self) -> List[Card]:
        """
        List of discarded cards, ordered by color and value.

        :return: the list of discarded cards, by lexicographic order. The order
            on the colors is the one specified in :attr:`cfg`.

        >>> from hanabython.Modules.Configuration import Configuration
        >>> discard_pile = DiscardPile(Configuration.STANDARD)
        >>> discard_pile.receive(Card('B3'))
        >>> discard_pile.receive(Card('R4'))
        >>> discard_pile.receive(Card('B1'))
        >>> discard_pile.list_reordered
        [<Card: B1>, <Card: B3>, <Card: R4>]
        """
        ordered = []
        for i, c in enumerate(self.cfg.colors):
            for j, v in enumerate(self.cfg.values):
                ordered.extend([Card(c, v)] * self.array[i, j])
        return ordered

    def receive(self, card) -> None:
        """
        Receive a card.

        :param card: the card discarded.

        Update the internal variables of the discard pile.

        >>> from hanabython.Modules.Configuration import Configuration
        >>> discard_pile = DiscardPile(Configuration.STANDARD)
        >>> discard_pile.receive(Card('B3'))
        >>> discard_pile.receive(Card('B2'))
        >>> discard_pile.receive(Card('B3'))
        >>> print(discard_pile)
        B2 B3 B3
        >>> print(discard_pile.not_discarded)
        [[3 1 0 2 1]
         [3 2 2 2 1]
         [3 2 2 2 1]
         [3 2 2 2 1]
         [3 2 2 2 1]]
        >>> print(discard_pile.scorable)
        [[ True  True False False False]
         [ True  True  True  True  True]
         [ True  True  True  True  True]
         [ True  True  True  True  True]
         [ True  True  True  True  True]]
        >>> print(discard_pile.max_score_possible)
        22
        """
        self.chronological.append(card)
        i = self.cfg.i_from_c(card.c)
        j = self.cfg.i_from_v(card.v)
        self.array[i, j] += 1
        self.not_discarded[i, j] -= 1
        if self.not_discarded[i, j] == 0:
            self.scorable[i, j:] = False


if __name__ == '__main__':
    my_discard_pile = DiscardPile(Configuration.W_MULTICOLOR_SHORT)
    my_discard_pile.receive(Card('R3'))
    my_discard_pile.receive(Card('R3'))
    my_discard_pile.receive(Card('M1'))
    my_discard_pile.receive(Card('B4'))
    my_discard_pile.receive(Card('B1'))
    my_discard_pile.test_str()

    print('\nAll layout styles (colored version):')
    print('Compact chronological: ')
    print(my_discard_pile.colored_compact_chronological())
    print('\nCompact ordered: ')
    print(my_discard_pile.colored_compact_ordered())
    print('\nCompact factorized: ')
    print(my_discard_pile.colored_compact_factorized())
    print('\nMulti-line compact: ')
    print(my_discard_pile.colored_multi_line_compact())
    print('\nMulti-line: ')
    print(my_discard_pile.colored_multi_line())
    print('\nAs an array: ')
    print(my_discard_pile.colored_as_array())

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
from typing import List
from hanabython.Modules.Clue import Clue
from hanabython.Modules.Action import Action
from hanabython.Modules.Card import Card
from hanabython.Modules.Colored import Colored
from hanabython.Modules.Configuration import Configuration


class Player(Colored):
    """
    A player for Hanabi.

    :param name: the name of the player.

    To define a subclass, the only real requirement is to implement the function
    :meth:`choose_action`.

    >>> antoine = Player('Antoine')
    >>> print(antoine)
    Antoine
    """

    def __init__(self, name: str):
        self.name = name

    def colored(self) -> str:
        return self.name

    # *** Game start ***

    def receive_init(self, cfg: Configuration, player_names: List[str]) -> None:
        """
        Receive a message: the game starts.

        :param cfg: the configuration of the game.
        :param player_names: the names of the players, rotated so that this
            player corresponds to index 0.
        """
        pass

    def receive_begin_dealing(self) -> None:
        """
        Receive a message: the initial dealing of hands begins.
        """
        pass

    def receive_end_dealing(self) -> None:
        """
        Receive a message: the initial dealing of hands is over.

        The hands themselves are not communicated in this message. Drawing
        cards, including for the initial hands, is always handled by
        :meth:`receive_i_draw` or :meth:`receive_partner_draws`.
        """
        pass

    # *** Drawing cards ***

    def receive_i_draw(self) -> None:
        """
        Receive a message: this player tries to draw a card.

        A card is actually drawn only if the draw pile is not empty.
        """
        pass

    def receive_partner_draws(self, i_active: int, card: Card) -> None:
        """
        Receive a message: another player tries to draw a card.

        A card is actually drawn only if the draw pile is not empty.

        :param i_active: the position of the player who draws (relatively
            to this player).
        :param card: the card drawn.
        """
        pass

    # *** General methods about actions ***

    def receive_turn_begin(self) -> None:
        """
        Receive a message: the turn of the player begins.
        """
        pass

    # noinspection PyMethodMayBeStatic
    def choose_action(self) -> Action:
        """
        Choose an action.

        :return: the action chosen by the player.
        """
        pass

    def receive_action_legal(self) -> None:
        """
        Receive a message: the action chosen is legal.
        """
        pass

    def receive_action_illegal(self, s: str) -> None:
        """
        Receive a message: the action chosen is illegal.

        :param s: a message explaining why the action is illegal.
        """
        pass

    def receive_turn_finished(self) -> None:
        """
        Receive a message: the turn of the player is finished.
        """
        pass

    # *** Manage the 4 types of actions ***

    def receive_someone_throws(self, i_active: int, k: int,
                               card: Card) -> None:
        """
        Receive a message: a player throws (discards a card willingly).

        It is not necessary to check whether this action is legal: the
        :attr:`Game` will only send this message when it is the case.

        :param i_active: the position of the player who throws (relatively
            to this player).
        :param k: position of the card in the hand.
        :param card: the card thrown.
        """
        pass

    def receive_someone_plays_card(
        self, i_active: int, k: int, card: Card
    ) -> None:
        """
        Receive a message: a player tries to play a card on the board.

        This can be a success or a misfire.

        :param i_active: the position of the player who plays the card
            (relatively to this player).
        :param k: position of the card in the hand.
        :param card: the card played.
        """
        pass

    def receive_someone_clues(
        self, i_active: int, i_clued: int, clue: Clue, bool_list: List[bool]
    ) -> None:
        """
        Receive a message: a player gives a clue to another one.

        It is not necessary to check whether this action is legal: the
        :attr:`Game` will only send this message when it is the case.

        :param i_active: the position of the player who gives the clue
            (relatively to this player).
        :param i_clued: the position of the player who receives the clue
            (relatively to this player).
        :param clue: the clue.
        :param bool_list: a list of boolean that indicates what cards
            match the clue given.
        """
        pass

    def receive_someone_forfeits(self, i_active: int) -> None:
        """
        Receive a message: a player forfeits.

        :param i_active: the position of the player who forfeits
            (relatively to this player).
        """
        pass

    # *** End of game ***

    def receive_remaining_turns(self, remaining_turns: int) -> None:
        """
        Receive a message: the number of remaining turns is now known.

        This happens with the normal rule for end of game: as soon as the
        discard pile is empty, we know how many turns are left. N.B.:
        the word  "turn" means that one player gets to play (not all of them).

        :param remaining_turns: the number of turns left.
        """
        pass

    def receive_lose(self, score: int) -> None:
        """
        Receive a message: the game is lost (misfires or forfeit).

        :param score: the final score (0 in that case).
        """
        pass

    def receive_game_exhausted(self, score: int) -> None:
        """
        Receive a message: the game is over and is neither really lost
        (misfires, forfeit) nor a total victory (maximal score). Typically,
        this happens a bit after the deck ran out of cards (it depends on the
        end-of-game rule that is used).

        :param score: the final score.
        """
        pass

    def receive_win(self, score: int) -> None:
        """
        Receive a message: the game is won (total victory).

        :param score: the final score.
        """
        pass


if __name__ == '__main__':
    my_antoine = Player(name='Antoine')
    my_antoine.test_str()

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
import logging
from copy import copy
from typing import List
from hanabython.Modules.Card import Card
from hanabython.Modules.Clue import Clue
from hanabython.Modules.Colors import Colors
from hanabython.Modules.Colored import Colored
from hanabython.Modules.Configuration import Configuration
from hanabython.Modules.ConfigurationEmptyClueRule \
    import ConfigurationEmptyClueRule
from hanabython.Modules.ConfigurationEndRule import ConfigurationEndRule
from hanabython.Modules.PlayerHumanText import PlayerHumanText
from hanabython.Modules.PlayerPuppet import PlayerPuppet
from hanabython.Modules.Board import Board
from hanabython.Modules.DrawPile import DrawPile
from hanabython.Modules.DiscardPile import DiscardPile
from hanabython.Modules.Hand import Hand
from hanabython.Modules.Action import Action
from hanabython.Modules.ActionClue import ActionClue
from hanabython.Modules.ActionThrow import ActionThrow
from hanabython.Modules.ActionForfeit import ActionForfeit
from hanabython.Modules.ActionPlayCard import ActionPlayCard
from hanabython.Modules.Player import Player


class Game(Colored):
    """
    A game of Hanabi.

    :param cfg: the configuration.
    :param players: the list of players. They will play in this order, starting
        with the first player in this list.

    :var int n_players: the number of players.
    :var Board board: the board.
    :var DrawPile draw_pile: the draw pile.
    :var DiscardPile discard_pile: the discard pile.
    :var int n_clues: the number of clue chips that players currently have.
    :var int n_misfires: the number of misfires chips that players currently
        have.
    :var int hand_size: the initial size of the hands.
    :var list hands: a list of :class:`Hand` objects (in the same order as
        :attr:`players`).
    :var int remaining_turns: the number of remaining turns (once the draw pile
        is empty, in the normal rule for end of game). As long as the draw pile
        contains cards, this variable is `None`.
    :var bool b_lose: the game is lost.
    :var bool b_win: the game is won.
    :var int i_active: the index of the active player.
    :var Player active: the active player. It is automatically updated when
        :attr:`i_active` is updated.

    >>> game = Game(players=[PlayerHumanText('Antoine'),
    ...                      PlayerHumanText('Donald X')])
    """

    def __init__(self, players: List[Player],
                 cfg: Configuration = Configuration.STANDARD):
        logging.info('General initializations')
        # Parameters
        self.players = players
        self.cfg = cfg
        # Variables
        self.n_players = len(self.players)                  # type: int
        self.board = Board(cfg)                             # type: Board
        self.draw_pile = DrawPile(cfg)                      # type: DrawPile
        self.discard_pile = DiscardPile(cfg)                # type: DiscardPile
        self.n_clues = cfg.n_clues                          # type: int
        self.n_misfires = 0                                 # type: int
        self.hand_size = cfg.hand_size_rule.f(self.n_players)  # type: int
        self.hands = [Hand() for _ in self.players]         # type: List[Hand]
        self.remaining_turns = None                         # type: int
        self.b_lose = False                                 # type: bool
        self.b_win = False                                  # type: bool
        # Active player
        self.active = None                                  # type: Player
        self._i_active = None                               # type: int
        # Inform the players of the initialization
        for i, p in enumerate(self.players):
            p.receive_init(cfg=copy(self.cfg), player_names=(
                [self.players[j].name for j in range(i, self.n_players)]
                + [self.players[j].name for j in range(i)]
            ))

    # *** Utils ***

    @property
    def i_active(self):
        """
        Index of the active player.

        :return: this index is automatically set modulo the number of players.
        >>> game = Game(players=[PlayerHumanText('Antoine'),
        ...                      PlayerHumanText('Donald X'),
        ...                      PlayerHumanText('Uwe')])
        >>> game.i_active = 2
        >>> game.i_active += 1
        >>> print(game.i_active)
        0
        """
        return self._i_active

    @i_active.setter
    def i_active(self, value):
        """
        Updates also :attr:`active`.
        """
        self._i_active = value % self.n_players
        self.active = self.players[self._i_active]

    def rel(self, who: int, fro: int) -> int:
        """
        Relative position of a player from the point of view of another one.

        :param who: the player we talk about.
        :param fro: the player to whom we talk.

        :return: the relative position of :attr:`who` from the point of view
            of :attr:`fro`, i.e. :attr:`who` - :attr:`fro` (modulo
            :attr:`n_players`).

        >>> game = Game(players=[PlayerHumanText('Antoine'),
        ...                      PlayerHumanText('Donald X'),
        ...                      PlayerHumanText('Uwe')])
        >>> game.rel(who=1, fro=2)
        2
        """
        return (who - fro) % self.n_players

    # *** Strings ***

    def colored(self) -> str:
        return (self.cfg.colored()
                + '\nPlayers: %s.' % ', '.join([p.name for p in self.players]))

    # *** Main method: play the game ***

    #: Number of attempts that a player has to choose her action. If she
    #: provides illegal actions as many times, she is automatically considered
    #: to forfeit (and this issues a warning).
    ATTEMPTS_BEFORE_FORFEIT = 100

    def play(self) -> int:
        """
        Main method: play the game.

        Note: it is only possible to "play" once with a :class:`Game` object. If
        you want to launch a game with the same player, it is necessary
        to define a new :class:`Game`.

        :return: the final score of the game.
        """
        logging.info("Begin dealing.")
        self.i_active = -1
        self.deal()
        logging.info("The game begins.")
        while True:
            self.i_active += 1
            logging.info("Check game-exhaustion condition.")
            if self.check_game_exhausted():
                return self.game_exhausted()
            logging.info("%s's turn begins" % self.active.name)
            self.active.receive_turn_begin()
            logging.info("Ask %s for an action." % self.active.name)
            for _ in range(Game.ATTEMPTS_BEFORE_FORFEIT):
                action = self.active.choose_action()
                is_legal = self.execute_action(action)
                if is_legal:
                    break
            else:  # i.e. if all the attempts were without a legal action
                logging.warning(
                    "%s failed 100 times to choose an action. Automatic "
                    "forfeit is applied." % self.active.name)
                self.execute_action(ActionForfeit())
            logging.info("Inform %s that his/her turn is over."
                         % self.active.name)
            self.active.receive_turn_finished()
            logging.info("Check win-or-lose condition.")
            if self.b_win:
                return self.win()
            if self.b_lose:
                return self.lose()
            if self.board.score == self.discard_pile.max_score_possible:
                return self.game_exhausted()

    # *** Drawing cards ***

    def draw(self) -> None:
        """
        The active player draws a card.

        * Draw a card and put it in hand (unless the discard pile is empty).

        * If the discard pile becomes empty, launch countdown for end of game
          by setting variable :attr:`remaining_turns` to value
          :attr:`n_players` + 1. It will be decremented at the beginning of
          next player's turn (before testing the end-of-game condition).
          Cf. :meth:`check_game_exhausted`.

        >>> game = Game(players=[PlayerHumanText('Antoine'),
        ...                      PlayerHumanText('Donald X'),
        ...                      PlayerHumanText('Uwe')])
        >>> game.i_active = -1
        >>> for _ in range(50):
        ...     game.i_active += 1
        ...     game.draw()
        >>> [len(hand) for hand in game.hands]
        [17, 17, 16]
        >>> game.i_active += 1
        >>> print(game.draw())
        None
        >>> game.remaining_turns
        4
        """
        logging.debug('Draw the card from draw pile and put it in the hand '
                      '(unless the draw pile is empty).')
        card = self.draw_pile.give()
        if card is not None:
            self.hands[self.i_active].receive(card)
        logging.debug('In normal rule for end of game, check if the countdown'
                      'for end of game should be launched.')
        if (self.cfg.end_rule == ConfigurationEndRule.NORMAL
                and self.draw_pile.n_cards == 0
                and self.remaining_turns is None):
            self.remaining_turns = self.n_players + 1
        logging.debug('Inform the players that someone drew.')
        for i, p in enumerate(self.players):
            if i == self.i_active:
                p.receive_i_draw()
            else:
                p.receive_partner_draws(self.rel(self.i_active, i), copy(card))

    def deal(self) -> None:
        """
        Deal the initial hands.

        :attr:`i_active` should be -1 before dealing and will be -1 at the end
        (modulo the number of players).

        >>> game = Game(players=[PlayerHumanText('Antoine'),
        ...                      PlayerHumanText('Donald X'),
        ...                      PlayerHumanText('Uwe')])
        >>> game.i_active = -1
        >>> game.deal()
        >>> [len(hand) for hand in game.hands]
        [5, 5, 5]
        >>> game.i_active
        2
        """
        logging.debug('Inform the players that dealing begins.')
        for p in self.players:
            p.receive_begin_dealing()
        logging.debug('Deal cards.')
        for _ in range(self.n_players * self.hand_size):
            self.i_active += 1
            self.draw()
        logging.debug('Inform the players that dealing is over.')
        for p in self.players:
            p.receive_end_dealing()

    # *** Manage the 4 types of actions ***

    def execute_action(self, action: Action) -> bool:
        """
        Execute the action (by the active player).

        :param action: the action.

        :return: True iff the action is legal. If not, it will be necessary
            to choose another action.

        This method dispatches to the auxiliary methods :meth:`execute_clue`,
        :meth:`execute_forfeit`, :meth:`execute_play_card` and
        :meth:`execute_throw`. Each of these methods has the responsability to:

        * Check if the action is legal,

        * Inform the active player whether it is the case or not,

        * Perform the action,

        * Update the relevant variables, in particular :attr:`b_lose` and
          :attr:`b_win`.

        * Inform all players of the result of the action,

        * Make the active player draw a new card if necessary,

        * Return the boolean stating whether the action is legal.
        """
        if isinstance(action, ActionForfeit):
            return self.execute_forfeit()
        if isinstance(action, ActionThrow):
            return self.execute_throw(k=action.k)
        if isinstance(action, ActionPlayCard):
            return self.execute_play_card(k=action.k)
        if isinstance(action, ActionClue):
            return self.execute_clue(
                i_clued=(self.i_active + action.i) % self.n_players,
                clue=action.clue)

    def execute_forfeit(self) -> bool:
        """
        Execute the action: forfeit.

        :return: True (meaning that this action is always legal).

        >>> game = Game(players=[PlayerPuppet('Antoine'),
        ...                      PlayerPuppet('Donald X'),
        ...                      PlayerPuppet('Uwe')])
        >>> game.players[1].speak = True
        >>> game.i_active = 1
        >>> is_legal = game.execute_forfeit()
        Donald X: The action chosen is legal.
        Donald X: A player forfeits.
        Donald X: i_active = 0
        >>> is_legal
        True
        >>> game.b_lose
        True
        >>> game.i_active = 2
        >>> is_legal = game.execute_forfeit()
        Donald X: A player forfeits.
        Donald X: i_active = 1
        >>> is_legal
        True
        >>> game.b_lose
        True
        """
        logging.debug('Check legality: forfeit is always legal.')
        logging.debug('Inform the active player that it is legal.')
        self.active.receive_action_legal()
        logging.debug('Perform the action.')
        self.b_lose = True
        logging.debug('Inform all players of the result of the action.')
        for i, p in enumerate(self.players):
            p.receive_someone_forfeits(self.rel(self.i_active, i))
        return True

    def execute_throw(self, k: int) -> bool:
        """
        Execute the action: throw (= discard willingly).

        :param k: the index of the card in the active player's hand.

        :return: True iff the action is legal, i.e. except if players have
            all the clue chips.

        >>> import random
        >>> random.seed(0)
        >>> game = Game(players=[PlayerPuppet('Antoine'),
        ...                      PlayerPuppet('Donald X'),
        ...                      PlayerPuppet('Uwe')])
        >>> game.players[1].speak = True
        >>> game.i_active = 1
        >>> game.draw()
        Donald X: This player tries to draw a card.
        >>> is_legal = game.execute_throw(0)
        Donald X: The action chosen is illegal.
        Donald X: You cannot discard because you have all the clue chips.
        >>> is_legal
        False
        >>> game.n_clues = 3
        >>> game.i_active = 2
        >>> game.draw()
        Donald X: Another player tries to draw a card.
        Donald X: i_active = 1
        Donald X: card = Y4
        >>> is_legal = game.execute_throw(0)
        Donald X: A player throws (discards a card willingly).
        Donald X: i_active = 1
        Donald X: k = 0
        Donald X: card = Y4
        Donald X: Another player tries to draw a card.
        Donald X: i_active = 1
        Donald X: card = R3
        >>> is_legal
        True
        >>> game.n_clues
        4
        >>> print(game.discard_pile)
        Y4
        >>> print(game.hands[2])
        R3
        """
        logging.debug('Discard: Check legality and inform the active player.')
        if self.n_clues == self.cfg.n_clues:
            self.active.receive_action_illegal(
                'You cannot discard because you have all the clue chips.')
            return False
        self.active.receive_action_legal()
        logging.debug('Perform the throw action.')
        card = self.hands[self.i_active].give(k)
        self.discard_pile.receive(card)
        self.n_clues += 1
        logging.debug('Inform all players of the result of the action.')
        for i, p in enumerate(self.players):
            p.receive_someone_throws(self.rel(self.i_active, i), k, card)
        logging.debug('Draw a card')
        self.draw()
        return True

    def execute_play_card(self, k: int) -> bool:
        """
        Execute the action: try to play a card.

        :param k: the index of the card in the active player's hand.

        :return: True (meaning that this action is always legal).

        The action can fail, in the sense that it leads to a misfire, but it
        is legal anyway. If it leads to the last misfire, then the players lose:

        >>> import random
        >>> random.seed(0)
        >>> game = Game(players=[PlayerPuppet('Antoine'),
        ...                      PlayerPuppet('Donald X'),
        ...                      PlayerPuppet('Uwe')])
        >>> game.i_active = -1
        >>> game.deal()
        >>> game.i_active = 2
        >>> game.n_misfires = 2
        >>> print(game.hands[2])
        B4 W4 G5 W1 R3
        >>> game.players[1].speak = True
        >>> is_legal = game.execute_play_card(2)
        Donald X: A player tries to play a card on the board.
        Donald X: i_active = 1
        Donald X: k = 2
        Donald X: card = G5
        >>> is_legal
        True
        >>> print(game.board)  #doctest: +NORMALIZE_WHITESPACE
        B -         G -         R -         W -         Y -
        >>> print(game.discard_pile)
        G5
        >>> game.n_misfires
        3
        >>> game.b_lose
        True

        If the highest card in a color is played, then the players gain a clue:

        >>> import random
        >>> random.seed(0)
        >>> game = Game(players=[PlayerPuppet('Antoine'),
        ...                      PlayerPuppet('Donald X'),
        ...                      PlayerPuppet('Uwe')])
        >>> game.i_active = -1
        >>> game.deal()
        >>> for s in ['G1', 'G2', 'G3', 'G4']:
        ...     _ = game.board.try_to_play(card=Card(s))
        >>> game.n_clues = 3
        >>> game.i_active = 2
        >>> print(game.hands[2])
        B4 W4 G5 W1 R3
        >>> game.players[1].speak = True
        >>> is_legal = game.execute_play_card(2)
        Donald X: A player tries to play a card on the board.
        Donald X: i_active = 1
        Donald X: k = 2
        Donald X: card = G5
        Donald X: Another player tries to draw a card.
        Donald X: i_active = 1
        Donald X: card = G4
        >>> is_legal
        True
        >>> print(game.board)  #doctest: +NORMALIZE_WHITESPACE
        B -         G 1 2 3 4 5 R -         W -         Y -
        >>> game.n_clues
        4

        But players cannot gain a clue if they already have the maximum
        number of clues:

        >>> import random
        >>> random.seed(0)
        >>> game = Game(players=[PlayerPuppet('Antoine'),
        ...                      PlayerPuppet('Donald X'),
        ...                      PlayerPuppet('Uwe')])
        >>> game.i_active = -1
        >>> game.deal()
        >>> for s in ['G1', 'G2', 'G3', 'G4']:
        ...     _ = game.board.try_to_play(card=Card(s))
        >>> game.n_clues
        8
        >>> game.i_active = 2
        >>> print(game.hands[2])
        B4 W4 G5 W1 R3
        >>> game.players[1].speak = True
        >>> is_legal = game.execute_play_card(2)
        Donald X: A player tries to play a card on the board.
        Donald X: i_active = 1
        Donald X: k = 2
        Donald X: card = G5
        Donald X: Another player tries to draw a card.
        Donald X: i_active = 1
        Donald X: card = G4
        >>> is_legal
        True
        >>> print(game.board)  #doctest: +NORMALIZE_WHITESPACE
        B -         G 1 2 3 4 5 R -         W -         Y -
        >>> game.n_clues
        8

        If the card completes the board, then the players win the game.

        >>> import random
        >>> random.seed(0)
        >>> game = Game(players=[PlayerPuppet('Antoine'),
        ...                      PlayerPuppet('Donald X'),
        ...                      PlayerPuppet('Uwe')])
        >>> game.i_active = -1
        >>> game.deal()
        >>> for c in ['B', 'R', 'W', 'Y']:
        ...     for v in range(1, 6):
        ...         _ = game.board.try_to_play(card=Card(c + str(v)))
        >>> for s in ['G1', 'G2', 'G3', 'G4']:
        ...     _ = game.board.try_to_play(card=Card(s))
        >>> game.i_active = 2
        >>> print(game.hands[2])
        B4 W4 G5 W1 R3
        >>> game.players[1].speak = True
        >>> _ = game.execute_play_card(2)
        Donald X: A player tries to play a card on the board.
        Donald X: i_active = 1
        Donald X: k = 2
        Donald X: card = G5
        >>> print(game.board)  #doctest: +NORMALIZE_WHITESPACE
        B 1 2 3 4 5 G 1 2 3 4 5 R 1 2 3 4 5 W 1 2 3 4 5 Y 1 2 3 4 5
        >>> game.b_win
        True
        """
        logging.debug('Check legality: play a card is always legal.')
        logging.debug('Inform the active player that it is legal.')
        self.active.receive_action_legal()
        logging.debug('Perform the "play a card" action.')
        card = self.hands[self.i_active].give(k)
        success = self.board.try_to_play(card)
        if success:
            if card.v == self.cfg.highest[card.c]:
                self.n_clues = min(self.n_clues + 1, self.cfg.n_clues)
            if self.board.score == self.cfg.max_score:
                self.b_win = True
        else:
            self.discard_pile.receive(card)
            self.n_misfires += 1
            if self.n_misfires == self.cfg.n_misfires:
                self.b_lose = True
        logging.debug('Inform all players of the result of the action.')
        for i, p in enumerate(self.players):
            p.receive_someone_plays_card(
                self.rel(self.i_active, i), k, copy(card))
        if not self.b_lose and not self.b_win:
            logging.debug('Draw a card')
            self.draw()
        return True

    def execute_clue(self, i_clued: int, clue: Clue) -> bool:
        """
        Execute the action: give a clue.

        :param i_clued: the index of the player who receives the clue.
        :param clue: the clue.

        :return: True iff the action is legal.

        >>> import random
        >>> random.seed(0)
        >>> game = Game(players=[PlayerPuppet('Antoine'),
        ...                      PlayerPuppet('Donald X'),
        ...                      PlayerPuppet('Uwe')],
        ...             cfg=Configuration.W_MULTICOLOR)
        >>> game.i_active = -1
        >>> game.deal()
        >>> print(game.hands[2])
        G2 W1 W1 B1 Y4
        >>> game.players[1].speak = True
        >>> game.i_active = 1
        >>> game.n_clues = 0
        >>> game.execute_clue(2, Clue(1))
        Donald X: The action chosen is illegal.
        Donald X: You cannot give a clue because you have do not have any clue \
chip.
        False
        >>> game.n_clues = 3
        >>> game.execute_clue(1, Clue(1))
        Donald X: The action chosen is illegal.
        Donald X: You cannot give a clue to yourself.
        False
        >>> game.execute_clue(2, Clue(6))
        Donald X: The action chosen is illegal.
        Donald X: This value does not exist: 6.
        False
        >>> game.execute_clue(2, Clue(Colors.COLORLESS))
        Donald X: The action chosen is illegal.
        Donald X: This color is not in the deck: C.
        False
        >>> game.execute_clue(2, Clue(Colors.MULTICOLOR))
        Donald X: The action chosen is illegal.
        Donald X: You cannot clue this color: M.
        False
        >>> game.execute_clue(2, Clue(3))
        Donald X: The action chosen is illegal.
        Donald X: You cannot give this clue because it does not correspond to \
any card.
        False
        >>> game.execute_clue(2, Clue(1))
        Donald X: The action chosen is legal.
        Donald X: A player gives a clue to another one.
        Donald X: i_active = 0
        Donald X: i_clued = 1
        Donald X: clue = 1
        Donald X: bool_list = [False, True, True, True, False]
        True
        >>> game.n_clues
        2
        """
        logging.debug('Check legality and inform the active player.')
        if self.n_clues == 0:
            self.active.receive_action_illegal(
                'You cannot give a clue because you have do not have any clue '
                'chip.')
            return False
        if i_clued == self.i_active:
            self.active.receive_action_illegal(
                'You cannot give a clue to yourself.')
            return False
        if (clue.category == Clue.VALUE
                and clue.x not in self.cfg.values):
            self.active.receive_action_illegal(
                'This value does not exist: %s.' % clue.x)
            return False
        if clue.category == Clue.COLOR and clue.x not in self.cfg.colors:
            self.active.receive_action_illegal(
                'This color is not in the deck: %s.' % clue.x)
            return False
        if clue.category == Clue.COLOR and not clue.x.is_cluable:
            self.active.receive_action_illegal(
                'You cannot clue this color: %s.' % clue.x.colored())
            return False
        bool_list = self.hands[i_clued].match(clue)
        if self.cfg.empty_clue_rule == ConfigurationEmptyClueRule.FORBIDDEN:
            if not any(bool_list):
                self.active.receive_action_illegal(
                    'You cannot give this clue because it does not correspond '
                    'to any card.')
                return False
        logging.debug('Inform the active player that the action is legal.')
        self.active.receive_action_legal()
        logging.debug('Perform the clue action.')
        self.n_clues -= 1
        logging.debug('Inform all players of the result of the action.')
        for i, p in enumerate(self.players):
            p.receive_someone_clues(
                self.rel(self.i_active, i), self.rel(i_clued, i), copy(clue),
                copy(bool_list))
        return True

    # *** End of game ***

    def check_game_exhausted(self) -> bool:
        """
        Check if the game is exhausted.

        Typically, the game end by exhaustion a bit after the deck ran out of
        cards (the exact moment depends on the end-of-game rule that is used).

        This method is called at the beginning of each player's turn.

        We do not check here whether the current score is equal to the maximum
        score still possible (considering what is discarded), which would also
        end the game. This verification is done in :meth:`play`.

        :return: True iff the game must end.

        If the normal end-of-game rule is used, and :attr:`remaining_turns`
        is an integer: it is updated, then the end-of-game condition is
        checked.

        >>> game = Game(players=[PlayerPuppet('Antoine'),
        ...                      PlayerPuppet('Donald X'),
        ...                      PlayerPuppet('Uwe')])
        >>> game.players[1].speak = True
        >>> game.remaining_turns = 1
        >>> # Here, previous player would play her turn.. Then...
        >>> game.check_game_exhausted()
        Donald X: The number of remaining turns is now known.
        Donald X: remaining_turns = 0
        True
        >>> print(game.remaining_turns)
        0

        If the normal end-of-game rule is used, and :attr:`remaining_turns`
        is `None`: it is not updated (the final countdown has not started).

        >>> game = Game(players=[PlayerPuppet('Antoine'),
        ...                      PlayerPuppet('Donald X'),
        ...                      PlayerPuppet('Uwe')])
        >>> game.players[1].speak = True
        >>> game.check_game_exhausted()
        False
        >>> print(game.remaining_turns)
        None

        If "Crowning piece" rule is used: if the active player has no card in
        hand, the game is over.

        >>> game = Game(
        ...     players=[PlayerPuppet('Antoine'),
        ...              PlayerPuppet('Donald X'),
        ...              PlayerPuppet('Uwe')],
        ...     cfg=Configuration(end_rule=ConfigurationEndRule.CROWNING_PIECE)
        ... )
        >>> game.players[1].speak = True
        >>> game.i_active = 1
        >>> game.check_game_exhausted()
        True
        >>> game.hands[1].receive(Card('B1'))
        >>> game.check_game_exhausted()
        False
        """
        if self.cfg.end_rule == ConfigurationEndRule.NORMAL:
            if self.remaining_turns is not None:
                self.remaining_turns -= 1
                for p in self.players:
                    p.receive_remaining_turns(self.remaining_turns)
                if self.remaining_turns == 0:
                    return True
        elif self.cfg.end_rule == ConfigurationEndRule.CROWNING_PIECE:
            if len(self.hands[self.i_active]) == 0:
                return True
        return False

    def lose(self) -> int:
        """
        Lose the game (forfeit or too many misfires).

        Inform the players.

        :return: the final score, i.e. 0.

        >>> game = Game(players=[PlayerPuppet('Antoine'),
        ...                      PlayerPuppet('Donald X'),
        ...                      PlayerPuppet('Uwe')])
        >>> game.players[1].speak = True
        >>> score = game.lose()
        Donald X: The game is lost.
        Donald X: score = 0
        >>> score
        0
        """
        for i, p in enumerate(self.players):
            p.receive_lose(score=0)
        return 0

    def game_exhausted(self) -> int:
        """
        The game is exhausted.

        Inform the players. The game is over and is neither really lost
        (misfires, forfeit) nor a total victory (maximal score). Typically,
        this happens a bit after the deck ran out of cards (it depends on the
        end-of-game rule that is used).

        :return: the final score.

        >>> game = Game(players=[PlayerPuppet('Antoine'),
        ...                      PlayerPuppet('Donald X'),
        ...                      PlayerPuppet('Uwe')])
        >>> _ = game.board.try_to_play(Card('B1'))
        >>> game.players[1].speak = True
        >>> score = game.game_exhausted()
        Donald X: The game is exhausted.
        Donald X: score = 1
        >>> score
        1
        """
        for i, p in enumerate(self.players):
            p.receive_game_exhausted(score=self.board.score)
        return self.board.score

    def win(self) -> int:
        """
        Win the game (maximum score).

        Inform the players.

        :return: the final score.

        >>> game = Game(players=[PlayerPuppet('Antoine'),
        ...                      PlayerPuppet('Donald X'),
        ...                      PlayerPuppet('Uwe')])
        >>> for c in ['B', 'G', 'R', 'W', 'Y']:
        ...     for v in range(1, 6):
        ...         _ = game.board.try_to_play(card=Card(c + str(v)))
        >>> game.players[1].speak = True
        >>> score = game.win()
        Donald X: The game is won.
        Donald X: score = 25
        >>> score
        25
        """
        for i, p in enumerate(self.players):
            p.receive_win(score=self.board.score)
        return self.board.score


if __name__ == '__main__':

    log_file = 'Game.log'
    open(log_file, 'w').close()  # Flush the contents of the file
    logging.basicConfig(
        filename=log_file,
        format='%(levelname)s - %(module)s - %(message)s',
        level=logging.DEBUG
    )

    # from hanabython.Modules.ConfigurationDeck import ConfigurationDeck
    # from hanabython.Modules.ConfigurationColorContents \
    #     import ConfigurationColorContents
    # cfg_test = Configuration(deck=ConfigurationDeck(contents=[
    #     (Colors.BLUE, ConfigurationColorContents([3, 2, 2])),
    #     (Colors.RED, ConfigurationColorContents([3, 2, 2])),
    # ]), end_rule=ConfigurationEndRule.CROWNING_PIECE)

    fanfan = PlayerHumanText(name='Fanfan')
    emilie = PlayerHumanText(name='Emilie')
    pek = PlayerHumanText(name='PEK')
    game = Game([fanfan, emilie])
    # game.test_str()
    game.play()

    # import doctest
    # doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
from typing import List
from hanabython.Modules.Colors import Colors
from hanabython.Modules.Clue import Clue
from hanabython.Modules.Colored import Colored
from hanabython.Modules.Configuration import Configuration
from hanabython.Modules.CardPublic import CardPublic


class HandPublic(Colored, list):
    """
    The "public" part of a hand.

    An object of this class represents what is known by all players, including
    the owner of the hand.

    We use the same convention as in Board Game Arena: newest cards are on the
    left (i.e. at the beginning of the list) and oldest cards are on the right
    (i.e. at the end of the list).

    Basically, a HandPublic is a list of CardPublic objects.

    :param cfg: the configuration of the game.
    :param n_cards: the number of cards in the hand. N.B.: this parameter
        is mostly used for examples and tests. In contrast, at the beginning of
        a game, the hand should be initialized with 0 cards, because cards will
        be given one by one to the players during the initial dealing of hands.

    >>> hand = HandPublic(cfg=Configuration.STANDARD, n_cards=4)
    >>> print(hand)
    BGRWY 12345, BGRWY 12345, BGRWY 12345, BGRWY 12345
    """
    def __init__(self, cfg: Configuration, n_cards: int = 0):
        super().__init__()
        self.cfg = cfg
        for i in range(n_cards):
            self.receive()

    def colored(self) -> str:
        return ', '.join([card.colored() for card in self])

    def receive(self) -> None:
        """
        Receive a card.

        An unknown card is added on the left, i.e. at the beginning of the list.

        >>> hand = HandPublic(cfg=Configuration.STANDARD, n_cards=4)
        >>> hand.match(clue=Clue(5), bool_list=[True, True, False, False])
        >>> print(hand)  #doctest: +NORMALIZE_WHITESPACE
        BGRWY 5  ,   BGRWY 5  , BGRWY 1234 , BGRWY 1234
        >>> hand.receive()
        >>> print(hand)  #doctest: +NORMALIZE_WHITESPACE
        BGRWY 12345,   BGRWY 5  ,   BGRWY 5  , BGRWY 1234 , BGRWY 1234
        """
        self.insert(0, CardPublic(self.cfg))

    def give(self, k: int) -> None:
        """
        Give a card.

        :param k: the position of the card in the hand (0 = newest).

        The card is simply suppressed from the hand.

        >>> hand = HandPublic(cfg=Configuration.STANDARD, n_cards=4)
        >>> hand.match(clue=Clue(5), bool_list=[False, True, False, False])
        >>> hand.match(clue=Clue(4), bool_list=[True, False, False, False])
        >>> print(hand)  #doctest: +NORMALIZE_WHITESPACE
        BGRWY 4  ,   BGRWY 5  ,  BGRWY 123 ,  BGRWY 123
        >>> hand.give(1)
        >>> print(hand)  #doctest: +NORMALIZE_WHITESPACE
        BGRWY 4  ,  BGRWY 123 ,  BGRWY 123
        """
        self.pop(k)

    def match(self, clue: Clue, bool_list: List[bool]):
        """
        React to a clue

        :param clue: the clue.
        :param bool_list: a list of booleans. The `i`-th coefficient is
            `True` iff the `i`-th card of the hand matches the clue given.

        Updates the internal variables of the hand.

        >>> hand = HandPublic(cfg=Configuration.STANDARD, n_cards=4)
        >>> hand.match(clue=Clue(3), bool_list=[False, True, False, False])
        >>> print(hand)  #doctest: +NORMALIZE_WHITESPACE
        BGRWY 1245 ,   BGRWY 3  , BGRWY 1245 , BGRWY 1245
        >>> hand.match(clue=Clue(Colors.RED),
        ...            bool_list=[False, True, False, False])
        >>> print(hand)  #doctest: +NORMALIZE_WHITESPACE
        BGWY 1245 ,     R3     ,  BGWY 1245 ,  BGWY 1245
        """
        for i, card in enumerate(self):
            card.match(clue=clue, b=bool_list[i])


if __name__ == '__main__':
    my_hand = HandPublic(cfg=Configuration(), n_cards=4)
    my_hand.test_str()

    print("\nLet's give some clues: ")
    print(my_hand.colored())
    my_hand.match(clue=Clue(Colors.RED),
                  bool_list=[True, False, True, False, False])
    print(my_hand.colored())
    my_hand.match(clue=Clue(Colors.BLUE),
                  bool_list=[False, True, False, False, False])
    print(my_hand.colored())
    my_hand.match(clue=Clue(3),
                  bool_list=[True, False, False, True, False])
    print(my_hand.colored())

    print("\nGive card in position 2: ")
    my_hand.give(2)
    print(my_hand.colored())
    print("Receive a new card: ")
    my_hand.receive()
    print(my_hand.colored())

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
from typing import List
from hanabython.Modules.Clue import Clue
from hanabython.Modules.Card import Card
from hanabython.Modules.StringUtils import uncolor, title
from hanabython.Modules.Configuration import Configuration
from hanabython.Modules.ConfigurationEndRule import ConfigurationEndRule
from hanabython.Modules.Board import Board
from hanabython.Modules.DiscardPile import DiscardPile
from hanabython.Modules.DrawPilePublic import DrawPilePublic
from hanabython.Modules.DrawPile import DrawPile
from hanabython.Modules.Hand import Hand
from hanabython.Modules.HandPublic import HandPublic
from hanabython.Modules.StringAnsi import StringAnsi
from hanabython.Modules.Player import Player


class PlayerBase(Player):
    """
    A player for Hanabi with basic features.

    This class is meant to serve as a mother class for most AIs and interface
    for human players. It provides all basic features, such as keeping
    track of the number of cards in the draw pile, the cards in the other
    players' hands, the clues given, etc.

    Note that all the variables are "personal" to this player: the :class:`Game`
    does not share access to its internal variables with the players.

    Note also that most methods are not supposed to work before
    :meth:`receive_init` is run at least once, which initializes all the
    variables for a new game.

    :param str name: the name of the player.

    :var list player_names: a list of strings, each with a player's name. By
        convention, the list is always rotated to that this player has
        position 0, the next player has position 1, etc.
    :var int n_players: the number of players.
    :var Configuration cfg: the configuration of the game.
    :var Board board: the board.
    :var DrawPilePublic draw_pile: the draw pile.
    :var DiscardPile discard_pile: the discard pile.
    :var int n_clues: the number of clues left.
    :var int n_misfires: the number of misfires (initially 0).
    :var int hand_size: the initial hand size.
    :var list hands: a list of Hand objects. The hand in position 0,
        corresponding to this player, is never updated because the player does
        not know what she has.
    :var list hands_public: a list of HandPublic objects. This allow the player
        to keep track, not only of her own clues, but also of the clues received
        by her partners.
    :var int remaining_turns: the number of remaining turns (once the draw pile
        is empty, in the normal rule for end of game). As long as the draw pile
        contains cards, this variable is `None`.
    :var str recent_events: things that happened "recently" (typically, since
        this player's last turn). Subclasses may typically print and/or empty
        this variable from time to time. Cf. :meth:`log`.
    :var bool dealing_is_ongoing: True only during the initial dealing of
        hands. Avoid useless verbose messages in :attr:`recent events`.
        Cf. :meth:`log`.
    :var int display_width: the width of the display on the terminal (in number
        of characters).

    >>> antoine = PlayerBase(name='Antoine')
    """
    def __init__(self, name: str):
        super().__init__(name)
        self.player_names = None        # type: List[str]
        self.n_players = None           # type: int
        self.cfg = None                 # type: Configuration
        self.board = None               # type: Board
        self.draw_pile = None           # type: DrawPilePublic
        self.discard_pile = None        # type: DiscardPile
        self.n_clues = None             # type: int
        self.n_misfires = None          # type: int
        self.hand_size = None           # type: int
        self.hands = None               # type: List[Hand]
        self.hands_public = None        # type: List[HandPublic]
        self.remaining_turns = None     # type: int
        self.dealing_is_ongoing = None  # type: bool
        self.recent_events = None       # type: str
        self.display_width = None       # type: int

    # *** String functions ***

    def __repr__(self) -> str:
        try:
            repr_width = (self.cfg.n_values + 1) * 2 * self.cfg.n_colors - 1
        except AttributeError:
            repr_width = 60
        return (
            '<PlayerBase:\n'
            + ''.join([
                title(attr, repr_width) + '\n'
                + str(self.__getattribute__(attr)) + '\n'
                for attr in sorted(self.__dict__.keys())
            ])
            + '>'
        )

    def colored(self) -> str:
        if self.cfg is None:
            return super().colored()
        # noinspection PyListCreation
        lines = []
        lines.append(title('Recent Events', self.display_width))
        lines.append(self.recent_events)
        lines.append(title('Hands', self.display_width))
        lines.append(self.colored_hands())
        lines.append(title('Board', self.display_width))
        lines.append(self.board.colored_fixed_space())
        lines.append(title('Discard Pile', self.display_width))
        lines.append(self.discard_pile.colored_compact_factorized())
        lines.append(title('Status', self.display_width))
        draw_line = 'Draw pile: %s.' % self.draw_pile.colored()
        if (self.cfg.end_rule == ConfigurationEndRule.NORMAL
                and self.remaining_turns is not None):
            draw_line += ' %s turns remaining!' % self.remaining_turns
        lines.append(draw_line)
        lines.append(
            (StringAnsi.BLUE + '%s' + StringAnsi.RESET
             + ' clues left (out of %s). '
             + StringAnsi.RED + '%s' + StringAnsi.RESET
             + ' misfires (out of %s).')
            % (self.n_clues, self.cfg.n_clues,
               self.n_misfires, self.cfg.n_misfires)
        )
        return '\n'.join(lines)

    def colored_hands(self) -> str:
        """
        A string used to display the hands of all players.

        :return: the string (whose width is usually :attr:`display_width`,
            except maybe in the end of game when the hands are shorter).

        >>> antoine = PlayerBase('Antoine')
        >>> antoine.demo_game()
        >>> from hanabython.Modules.StringUtils import uncolor
        >>> print(uncolor(antoine.colored_hands()))  \
#doctest: +NORMALIZE_WHITESPACE
        Antoine
        BGRWY 12345, BGRWY 2345 ,   BGRWY 1  ,   BGRWY 1  , BGRWY 2345
        <BLANKLINE>
        Donald X
            Y2     ,     R1     ,     R3     ,     G3     ,     Y4
        BGRWY 2345 ,   BGRWY 1  , BGRWY 2345 , BGRWY 2345 , BGRWY 2345
        <BLANKLINE>
        Uwe
            G4     ,     B4     ,     W4     ,     G5     ,     W1
        BGRWY 12345, BGRWY 12345, BGRWY 12345, BGRWY 12345, BGRWY 12345
        """
        # noinspection PyListCreation
        lines = []
        lines.append(self.name)
        lines.append(self.hands_public[0].colored())
        for i, name in enumerate(self.player_names):
            if i == 0:
                continue
            lines.append('')
            lines.append(name)
            lines.append(', '.join([
                self._large_card_color(card) for card in self.hands[i]
            ]))
            lines.append(self.hands_public[i].colored())
        return '\n'.join(lines)

    # noinspection PyProtectedMember
    def _large_card(self, card: Card) -> str:
        """
        A string for a :class:`Card`, but as wide as a :attr:`CardPublic`.

        This works only once :attr:`cfg` is initialized, which is done via
        :meth:`receive_init`. Indeed, the configuration is needed to known
        the width.

        :param card: the card.

        :return: the large string.

        >>> from hanabython.Modules.Card import Card
        >>> from hanabython.Modules.CardPublic import CardPublic
        >>> antoine = PlayerBase('Antoine')
        >>> antoine.receive_init(cfg=Configuration.STANDARD,
        ...                      player_names=['Antoine', 'Donald X'])
        >>> print('<%s>' % antoine._large_card(Card('B5')))
        <    B5     >
        >>> print('<%s>' % CardPublic(antoine.cfg))
        <BGRWY 12345>
        """
        return uncolor(self._large_card_color(card))

    # noinspection PyProtectedMember
    def _large_card_color(self, card: Card) -> str:
        """
        Colored version of :meth:`_large_card`
        """
        width = self.cfg.n_colors + self.cfg.n_values + 1
        s = str(card)
        left = (width - len(s)) // 2
        right = width - len(s) - left
        return ' ' * left + card.colored() + ' ' * right

    # *** Internal log (for the player) ***

    def log(self, o: object) -> None:
        """
        Log events for the player.

        :param o: an object. The method adds `str(o)` to the variable
            :attr:`recent_events`, except during the initial dealing of cards
            (to avoid useless messages about each card dealt). Do not forget
            the end-of-line character when relevant (it is not added
            automatically).

        This is for the player herself: it is used, in particular, in the
        subclass :class:`PlayerHumanText` to inform the player of the most
        recent events in a relatively user-friendly form.

        N.B.: this is totally different from the use of the standard package
        ``logging``, which is essentially used for debugging purposes.

        >>> antoine = PlayerBase('Antoine')
        >>> antoine.log_init()
        >>> antoine.log('Something happens.\\n')
        >>> antoine.dealing_is_ongoing = True
        >>> antoine.log('Many useless messages.\\n')
        >>> antoine.dealing_is_ongoing = False
        >>> antoine.log('Something else happens.\\n')
        >>> print(antoine.recent_events)
        Something happens.
        Something else happens.
        <BLANKLINE>
        >>> antoine.log_forget()
        >>> antoine.log('Something new happens.')
        >>> print(antoine.recent_events)
        Something new happens.
        """
        if not self.dealing_is_ongoing:
            self.recent_events += str(o)

    def log_init(self) -> None:
        """
        Initialize the log process (at the beginning of a game).

        Empties :attr:`recent_events`.
        """
        self.recent_events = ''

    def log_forget(self) -> None:
        """
        Forget old events (during the game).

        Empties :attr:`recent_events`. In this base class, this method has the
        same implementation as :meth:`log_init`, but it could be different
        in some subclasses.
        """
        self.recent_events = ''

    # *** Game start ***

    def receive_init(self, cfg: Configuration, player_names: List[str]) -> None:
        """
        Initialize all the instance variables for a new game.

        >>> antoine = PlayerBase('Antoine')
        >>> antoine.receive_init(Configuration.STANDARD,
        ...                      player_names=['Antoine', 'Donald X'])
        >>> print(repr(antoine))  #doctest: +NORMALIZE_WHITESPACE
        <PlayerBase:
        ************************** board **************************
        B -         G -         R -         W -         Y -
        *************************** cfg ***************************
        Deck: normal.
        Number of clues: 8.
        Number of misfires: 3.
        Clues rule: empty clues are forbidden.
        End rule: normal.
        ******************* dealing_is_ongoing ********************
        False
        ********************** discard_pile ***********************
        No card discarded yet
        ********************** display_width **********************
        63
        ************************ draw_pile ************************
        50 cards left
        ************************ hand_size ************************
        5
        ************************** hands **************************
        [<Hand: >, <Hand: >]
        ********************** hands_public ***********************
        [<HandPublic: >, <HandPublic: >]
        ************************* n_clues *************************
        8
        *********************** n_misfires ************************
        0
        ************************ n_players ************************
        2
        ************************** name ***************************
        Antoine
        ********************** player_names ***********************
        ['Antoine', 'Donald X']
        ********************** recent_events **********************
        Configuration
        -------------
        Deck: normal.
        Number of clues: 8.
        Number of misfires: 3.
        Clues rule: empty clues are forbidden.
        End rule: normal.
        <BLANKLINE>
        ********************* remaining_turns *********************
        None
        >
        """
        self.player_names = player_names
        self.n_players = len(player_names)
        self.cfg = cfg
        self.board = Board(cfg)
        self.draw_pile = DrawPilePublic(cfg)
        self.discard_pile = DiscardPile(cfg)
        self.n_clues = cfg.n_clues
        self.n_misfires = 0
        self.hand_size = cfg.hand_size_rule.f(self.n_players)
        self.hands = [Hand() for _ in player_names]
        self.hands_public = [HandPublic(cfg) for _ in player_names]
        self.dealing_is_ongoing = False
        self.display_width = (
            self.cfg.n_colors + 3 + self.cfg.n_values) * self.hand_size - 2
        self.log_init()
        self.log('Configuration\n')
        self.log('-------------\n')
        self.log(self.cfg.colored())
        self.log('\n')

    def receive_begin_dealing(self) -> None:
        """
        The log is turned off to avoid having a message for each card dealt.
        Cf. :meth:`log`.

        >>> antoine = PlayerBase('Antoine')
        >>> antoine.receive_init(Configuration.STANDARD,
        ...                      player_names=['Antoine', 'Donald X'])
        >>> antoine.receive_begin_dealing()
        >>> antoine.dealing_is_ongoing
        True
        >>> antoine.receive_end_dealing()
        >>> antoine.dealing_is_ongoing
        False
        """
        self.dealing_is_ongoing = True

    def receive_end_dealing(self) -> None:
        """
        The log is turned back on. Cf. :meth:`log` and
        :meth:`receive_begin_dealing`.
        """
        self.dealing_is_ongoing = False
        # self.log('\nInitial hands' + '\n')
        # self.log('-------------' + '\n')
        # self.log(self.colored_hands() + '\n')
        self.log('\nFirst moves' + '\n')
        self.log('-----------' + '\n')
        self.log('The game begins.\n')

    # *** Drawing cards ***

    def receive_i_draw(self) -> None:
        """
        If there are cards in the draw pile, a card is drawn. There is one
        card less in :attr:`drawpile`, and one more in this player's hand in
        :attr:`hands_public`.

        >>> antoine = PlayerBase('Antoine')
        >>> antoine.receive_init(Configuration.STANDARD,
        ...                      player_names=['Antoine', 'Donald X'])
        >>> for _ in range(4):
        ...     antoine.receive_i_draw()
        >>> print(antoine.draw_pile)
        46 cards left
        >>> print(antoine.hands_public[0])
        BGRWY 12345, BGRWY 12345, BGRWY 12345, BGRWY 12345

        If there are no cards in the draw pile, nothing happens.

        >>> antoine = PlayerBase('Antoine')
        >>> antoine.receive_init(Configuration.STANDARD,
        ...                      player_names=['Antoine', 'Donald X'])
        >>> for _ in range(50):
        ...     antoine.receive_i_draw()
        >>> len(antoine.hands_public[0])
        50
        >>> print(antoine.draw_pile)
        No card left
        >>> antoine.receive_i_draw()
        >>> len(antoine.hands_public[0])
        50
        >>> print(antoine.draw_pile)
        No card left
        """
        if self.draw_pile.n_cards == 0:
            return
        self.draw_pile.give()
        self.hands_public[0].receive()
        self.log('%s draws a card.\n' % self.name)

    def receive_partner_draws(self, i_active: int, card: Card) -> None:
        """
        If there are cards in the draw pile, a card is drawn. There is one
        card less in :attr:`drawpile`, one more in the partner's hand in
        :attr:`hands_public`, and the actual card is added in the partner's
        hand in :attr:`hands`.

        >>> antoine = PlayerBase('Antoine')
        >>> antoine.receive_init(Configuration.STANDARD,
        ...                      player_names=['Antoine', 'Donald X'])
        >>> for s in ['B1', 'G3', 'Y1', 'W1', 'R5']:
        ...     antoine.receive_partner_draws(i_active=1, card=Card(s))
        >>> print(antoine.draw_pile)
        45 cards left
        >>> print(antoine.hands[1])
        R5 W1 Y1 G3 B1
        >>> print(antoine.hands_public[1])
        BGRWY 12345, BGRWY 12345, BGRWY 12345, BGRWY 12345, BGRWY 12345

        If there are no cards in the draw pile, nothing happens.

        >>> antoine = PlayerBase('Antoine')
        >>> antoine.receive_init(Configuration.STANDARD,
        ...                      player_names=['Antoine', 'Donald X'])
        >>> for _ in range(50):
        ...     antoine.receive_partner_draws(i_active=1, card=Card('B1'))
        >>> len(antoine.hands[1])
        50
        >>> len(antoine.hands_public[1])
        50
        >>> print(antoine.draw_pile)
        No card left
        >>> antoine.receive_i_draw()
        >>> len(antoine.hands[1])
        50
        >>> len(antoine.hands_public[1])
        50
        >>> print(antoine.draw_pile)
        No card left
        """
        if card is None:
            return
        self.draw_pile.give()
        self.hands[i_active].receive(card)
        self.hands_public[i_active].receive()
        self.log('%s draws %s.\n' % (
            self.player_names[i_active], card.colored()))

    # *** Manage the 4 types of actions ***

    def receive_someone_throws(
        self, i_active: int, k: int, card: Card
    ) -> None:
        """
        The card goes in the discard pile, and players regain a clue chip.

        >>> antoine = PlayerBase('Antoine')
        >>> antoine.receive_init(Configuration.STANDARD,
        ...                      player_names=['Antoine', 'Donald X'])
        >>> antoine.n_clues = 3
        >>> for s in ['B1', 'G3', 'Y1', 'W1', 'R5']:
        ...     antoine.receive_partner_draws(i_active=1, card=Card(s))
        >>> print(antoine.hands[1])
        R5 W1 Y1 G3 B1
        >>> antoine.receive_someone_throws(i_active=1, k=4, card=Card('B1'))
        >>> print(antoine.hands[1])
        R5 W1 Y1 G3
        >>> print(antoine.hands_public[1])
        BGRWY 12345, BGRWY 12345, BGRWY 12345, BGRWY 12345
        >>> print(antoine.discard_pile)
        B1
        >>> antoine.n_clues
        4
        """
        self.hands_public[i_active].give(k)
        if i_active != 0:
            self.hands[i_active].give(k)
        self.discard_pile.receive(card)
        self.n_clues += 1
        self.log('%s discards %s.\n' % (
            self.player_names[i_active], card.colored()))

    def receive_someone_plays_card(
        self, i_active: int, k: int, card: Card
    ) -> None:
        """
        If the action succeeds, the card goes on the board.

        >>> antoine = PlayerBase('Antoine')
        >>> antoine.receive_init(Configuration.STANDARD,
        ...                      player_names=['Antoine', 'Donald X'])
        >>> for s in ['B1', 'G3', 'Y1', 'W1', 'R5']:
        ...     antoine.receive_partner_draws(i_active=1, card=Card(s))
        >>> print(antoine.hands[1])
        R5 W1 Y1 G3 B1
        >>> antoine.receive_someone_plays_card(i_active=1, k=1, card=Card('W1'))
        >>> print(antoine.hands[1])
        R5 Y1 G3 B1
        >>> print(antoine.hands_public[1])
        BGRWY 12345, BGRWY 12345, BGRWY 12345, BGRWY 12345
        >>> print(antoine.board)  #doctest: +NORMALIZE_WHITESPACE
        B -         G -         R -         W 1         Y -

        If the action fails, the card goes in the discard pile and players get
        a misfire.

        >>> antoine = PlayerBase('Antoine')
        >>> antoine.receive_init(Configuration.STANDARD,
        ...                      player_names=['Antoine', 'Donald X'])
        >>> for s in ['B1', 'G3', 'Y1', 'W1', 'R5']:
        ...     antoine.receive_partner_draws(i_active=1, card=Card(s))
        >>> print(antoine.hands[1])
        R5 W1 Y1 G3 B1
        >>> antoine.receive_someone_plays_card(i_active=1, k=0, card=Card('R5'))
        >>> print(antoine.hands[1])
        W1 Y1 G3 B1
        >>> print(antoine.hands_public[1])
        BGRWY 12345, BGRWY 12345, BGRWY 12345, BGRWY 12345
        >>> print(antoine.board)  #doctest: +NORMALIZE_WHITESPACE
        B -         G -         R -         W -         Y -
        >>> print(antoine.discard_pile)
        R5
        >>> antoine.n_misfires
        1
        """
        self.hands_public[i_active].give(k)
        if i_active != 0:
            self.hands[i_active].give(k)
        success = self.board.try_to_play(card)
        if success:
            self.log('%s plays %s' % (
                self.player_names[i_active], card.colored()))
            if (card.v == self.cfg.highest[card.c]
                    and self.n_clues < self.cfg.n_clues):
                self.n_clues += 1
                self.log(' and regains a clue.\n')
            else:
                self.log('.\n')
        else:
            self.discard_pile.receive(card)
            self.n_misfires += 1
            self.log('%s tries to play %s and misfires.\n' % (
                self.player_names[i_active], card.colored()))

    def receive_someone_clues(
        self, i_active: int, i_clued: int, clue: Clue, bool_list: List[bool]
    ) -> None:
        """
        We remove a clue chip, and we update the clued player's hand in
        :attr:`hands_public`.

        >>> antoine = PlayerBase('Antoine')
        >>> antoine.receive_init(Configuration.STANDARD,
        ...                      player_names=['Antoine', 'Donald X'])
        >>> for s in ['B1', 'G3', 'Y1', 'W1', 'R5']:
        ...     antoine.receive_partner_draws(i_active=1, card=Card(s))
        >>> print(antoine.hands[1])
        R5 W1 Y1 G3 B1
        >>> antoine.n_clues
        8
        >>> antoine.receive_someone_clues(
        ...     i_active=0, i_clued=1, clue=Clue(1),
        ...     bool_list=[False, True, True, False, True])
        >>> print(antoine.hands_public[1])  #doctest: +NORMALIZE_WHITESPACE
        BGRWY 2345 ,   BGRWY 1  ,   BGRWY 1  , BGRWY 2345 ,   BGRWY 1
        >>> antoine.n_clues
        7
        """
        self.n_clues -= 1
        self.hands_public[i_clued].match(clue, bool_list)
        self.log(
            '%s clues %s about %s.\n'
            % (self.player_names[i_active], self.player_names[i_clued],
               clue.colored())
        )

    def receive_someone_forfeits(self, i_active: int) -> None:
        """
        We just log the event for the player.

        >>> antoine = PlayerBase('Antoine')
        >>> antoine.receive_init(Configuration.STANDARD,
        ...                      player_names=['Antoine', 'Donald X'])
        >>> antoine.log_forget()
        >>> antoine.receive_someone_forfeits(i_active=1)
        >>> print(antoine.recent_events)
        Donald X forfeits.
        <BLANKLINE>
        """
        self.log('%s forfeits.\n' % self.player_names[i_active])

    # *** End of game ***

    def receive_remaining_turns(self, remaining_turns: int) -> None:
        """
        We update :attr:`remaining_turns` and log the event for the player.

        >>> antoine = PlayerBase('Antoine')
        >>> antoine.receive_init(Configuration.STANDARD,
        ...                      player_names=['Antoine', 'Donald X'])
        >>> antoine.log_forget()
        >>> antoine.receive_remaining_turns(remaining_turns=2)
        >>> antoine.remaining_turns
        2
        >>> print(antoine.recent_events)
        2 turns remaining!
        <BLANKLINE>
        """
        self.remaining_turns = remaining_turns
        self.log('%s turns remaining!\n' % self.remaining_turns)

    def receive_lose(self, score: int) -> None:
        """
        We just log the event for the player.

        >>> antoine = PlayerBase('Antoine')
        >>> antoine.receive_init(Configuration.STANDARD,
        ...                      player_names=['Antoine', 'Donald X'])
        >>> antoine.log_forget()
        >>> antoine.receive_lose(score=0)
        >>> print(antoine.recent_events)
        Antoine's team loses.
        Score: 0.
        <BLANKLINE>
        """
        self.log("%s's team loses.\n" % self.name)
        self.log('Score: %s.\n' % score)

    def receive_game_exhausted(self, score: int) -> None:
        """
        We just log the event for the player.

        >>> antoine = PlayerBase('Antoine')
        >>> antoine.receive_init(Configuration.STANDARD,
        ...                      player_names=['Antoine', 'Donald X'])
        >>> antoine.log_forget()
        >>> antoine.receive_game_exhausted(score=23)
        >>> print(antoine.recent_events)
        Antoine's team has reached the end of the game.
        Score: 23.
        <BLANKLINE>
        """
        self.log("%s's team has reached the end of the game.\n" % self.name)
        self.log('Score: %s.\n' % score)

    def receive_win(self, score: int) -> None:
        """
        We just log the event for the player.

        >>> antoine = PlayerBase('Antoine')
        >>> antoine.receive_init(Configuration.STANDARD,
        ...                      player_names=['Antoine', 'Donald X'])
        >>> antoine.log_forget()
        >>> antoine.receive_win(score=25)
        >>> print(antoine.recent_events)
        Antoine's team wins!
        Score: 25.
        <BLANKLINE>
        """
        self.log("%s's team wins!\n" % self.name)
        self.log('Score: %s.\n' % score)

    # *** Demo ***

    def demo_game(self) -> None:
        """
        Mimic the beginning of a game.

        This method is meant to be used for tests and examples.

        >>> from hanabython.Modules.StringUtils import uncolor
        >>> antoine = PlayerBase('Antoine')
        >>> antoine.demo_game()
        >>> print(uncolor(antoine.recent_events))
        Configuration
        -------------
        Deck: normal.
        Number of clues: 8.
        Number of misfires: 3.
        Clues rule: empty clues are forbidden.
        End rule: normal.
        <BLANKLINE>
        First moves
        -----------
        The game begins.
        Antoine clues Donald X about 1.
        Donald X clues Antoine about 1.
        Uwe discards R3.
        Uwe draws G4.
        Antoine plays W1.
        Antoine draws a card.
        <BLANKLINE>
        """
        import random
        random.seed(0)
        cfg = Configuration.STANDARD
        draw_pile = DrawPile(cfg)
        self.receive_init(cfg=cfg, player_names=[self.name, 'Donald X', 'Uwe'])
        self.receive_begin_dealing()
        my_hand = Hand()
        for k in range(self.hand_size):
            my_hand.receive(card=draw_pile.give())
            self.receive_i_draw()
            for i in range(1, self.n_players):
                self.receive_partner_draws(i_active=i, card=draw_pile.give())
        self.receive_end_dealing()
        self.receive_someone_clues(
            i_active=0, i_clued=1, clue=Clue(1),
            bool_list=self.hands[1].match(Clue(1)))
        self.receive_someone_clues(
            i_active=1, i_clued=0, clue=Clue(1),
            bool_list=my_hand.match(Clue(1)))
        self.receive_someone_throws(i_active=2, k=4, card=self.hands[2][4])
        self.receive_partner_draws(i_active=2, card=draw_pile.give())
        self.receive_someone_plays_card(i_active=0, k=1, card=my_hand[1])
        my_hand.receive(card=draw_pile.give())
        self.receive_i_draw()


if __name__ == '__main__':
    my_antoine = PlayerBase(name='Antoine')
    my_antoine.test_str()
    my_antoine.receive_init(Configuration.STANDARD,
                            player_names=['Antoine', 'Donald X'])
    my_antoine.test_str()

    print("""
    *******************************************************
    *                                                     *
    * Pretend that the beginning of the game is played... *
    *                                                     *
    *******************************************************
    """)
    my_antoine.demo_game()
    my_antoine.test_str()

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
from hanabython.Modules.Color import Color
from hanabython.Modules.StringAnsi import StringAnsi


class ColorColorless(Color):

    def match(self, clue_color: Color) -> bool:
        """
        Colorless matches no color clue.

        >>> colorless = ColorColorless(name='Colorless', symbol='C',
        ...                            print_color=StringAnsi.MAGENTA)
        >>> brown = Color(name='Brown', symbol='N',
        ...               print_color=StringAnsi.BROWN)
        >>> colorless.match(clue_color=brown)
        False
        """
        return False

    @property
    def is_cluable(self):
        """
        :return: False. It is not allowed to give "colorless" as a clue.
        """
        return False


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
from typing import List
from hanabython.Modules.Clue import Clue
from hanabython.Modules.Action import Action
from hanabython.Modules.ActionForfeit import ActionForfeit
from hanabython.Modules.Card import Card
from hanabython.Modules.Configuration import Configuration
from hanabython.Modules.Player import Player
from hanabython.Modules.StringUtils import uncolor


class PlayerPuppet(Player):
    """
    A player for Hanabi that serves only for testing purposes.

    :param speak: if True, then each time this player receives a message, she
        prints a acknowledgement.

    :var Action next_action: this variable makes it possible to control
        this player's action.

    >>> from hanabython.Modules.ActionThrow import ActionThrow
    >>> antoine = PlayerPuppet('Antoine', speak=True)
    >>> antoine.next_action = ActionThrow(k=4)
    >>> _ = antoine.choose_action()
    Antoine: Choose an action
    Antoine: action = Discard card in position 5
    """

    def __init__(self, name, speak=False):
        super().__init__(name)
        self.next_action = ActionForfeit()              # type: Action
        self.dealing_is_ongoing = False                 # type: bool
        self.speak = speak                              # type: bool

    def ack(self, o):
        if self.speak and not self.dealing_is_ongoing:
            print(uncolor('%s: %s' % (self.name, o)))

    # *** Game start ***

    def receive_init(self, cfg: Configuration, player_names: List[str]) -> None:
        self.ack('The game starts')
        self.ack('cfg = %s' % cfg)
        self.ack('player_names = %s' % player_names)

    def receive_begin_dealing(self) -> None:
        self.ack('the initial dealing of hands begins.')
        self.dealing_is_ongoing = True

    def receive_end_dealing(self) -> None:
        self.dealing_is_ongoing = False
        self.ack('The initial dealing of hands is over.')

    # *** Drawing cards ***

    def receive_i_draw(self) -> None:
        self.ack('This player tries to draw a card.')

    def receive_partner_draws(self, i_active: int, card: Card) -> None:
        self.ack('Another player tries to draw a card.')
        self.ack('i_active = %s' % i_active)
        self.ack('card = %s' % card)

    # *** General methods about actions ***

    def receive_turn_begin(self) -> None:
        self.ack('The turn of the player begins.')

    # noinspection PyMethodMayBeStatic
    def choose_action(self) -> Action:
        """
        :return: the value of :attr:`next_action`
        """
        self.ack('Choose an action')
        self.ack('action = %s' % self.next_action)
        return self.next_action

    def receive_action_legal(self) -> None:
        self.ack('The action chosen is legal.')

    def receive_action_illegal(self, s: str) -> None:
        self.ack('The action chosen is illegal.')
        self.ack(s)

    def receive_turn_finished(self) -> None:
        self.ack('The action of the player is finished.')

    # *** Manage the 4 types of actions ***

    def receive_someone_throws(self, i_active: int, k: int,
                               card: Card) -> None:
        self.ack('A player throws (discards a card willingly).')
        self.ack('i_active = %s' % i_active)
        self.ack('k = %s' % k)
        self.ack('card = %s' % card)

    def receive_someone_plays_card(
        self, i_active: int, k: int, card: Card
    ) -> None:
        self.ack('A player tries to play a card on the board.')
        self.ack('i_active = %s' % i_active)
        self.ack('k = %s' % k)
        self.ack('card = %s' % card)

    def receive_someone_clues(
        self, i_active: int, i_clued: int, clue: Clue, bool_list: List[bool]
    ) -> None:
        self.ack('A player gives a clue to another one.')
        self.ack('i_active = %s' % i_active)
        self.ack('i_clued = %s' % i_clued)
        self.ack('clue = %s' % clue)
        self.ack('bool_list = %s' % bool_list)

    def receive_someone_forfeits(self, i_active: int) -> None:
        self.ack('A player forfeits.')
        self.ack('i_active = %s' % i_active)

    # *** End of game ***

    def receive_remaining_turns(self, remaining_turns: int) -> None:
        self.ack('The number of remaining turns is now known.')
        self.ack('remaining_turns = %s' % remaining_turns)

    def receive_lose(self, score: int) -> None:
        self.ack('The game is lost.')
        self.ack('score = %s' % score)

    def receive_game_exhausted(self, score: int) -> None:
        self.ack('The game is exhausted.')
        self.ack('score = %s' % score)

    def receive_win(self, score: int) -> None:
        self.ack('The game is won.')
        self.ack('score = %s' % score)


if __name__ == '__main__':
    my_antoine = PlayerPuppet(name='Antoine')
    my_antoine.test_str()

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
from typing import Union
from hanabython.Modules.Colored import Colored
from random import shuffle
from hanabython.Modules.Configuration import Configuration
from hanabython.Modules.Card import Card


class DrawPile(Colored, list):
    """
    The draw pile of a game of Hanabi.

    :param cfg: the configuration of the game.

    At initialization, the draw pile is generated with the parameters in
    :attr:`cfg`, then it is shuffled.

    Basically, a DrawPile is a list of cards. The top of the pile, where cards
    are drawn, is represented by the end of the list (not that we care much,
    but it could have an influence someday in some not-yet-implemented
    non-official variants).

    >>> from hanabython.Modules.Configuration import Configuration
    >>> draw_pile = DrawPile(Configuration.STANDARD)
    """

    def __init__(self, cfg: Configuration):
        super().__init__()
        self.cfg = cfg
        for i, c in enumerate(cfg.colors):
            for j, v in enumerate(cfg.values):
                self.extend([Card(c, v)] * cfg.deck[c][j])
        shuffle(self)

    def colored(self) -> str:
        return '[' + ', '.join([card.colored() for card in self]) + ']'

    @property
    def n_cards(self) -> int:
        """
        Number of cards in the pile.

        :return: the number of cards.

        >>> from hanabython.Modules.Configuration import Configuration
        >>> draw_pile = DrawPile(Configuration.STANDARD)
        >>> draw_pile.n_cards
        50
        """
        return len(self)

    def give(self) -> Union[Card, None]:
        """
        Give the card from the top of pile.

        :return: the card drawn. If the pile is empty, return None.

        >>> from hanabython.Modules.Configuration import Configuration
        >>> draw_pile = DrawPile(cfg=Configuration.STANDARD)
        >>> card = draw_pile.give()
        >>> type(card)
        <class 'hanabython.Modules.Card.Card'>
        >>> while draw_pile.n_cards >= 1:
        ...     _ = draw_pile.give()
        >>> print(draw_pile.give())
        None
        """
        if self.n_cards == 0:
            return None
        return self.pop()


if __name__ == '__main__':
    my_draw_pile = DrawPile(cfg=Configuration.W_MULTICOLOR_SHORT)
    my_draw_pile.test_str()

    print('\nDraw a card: ')
    print(my_draw_pile.colored())
    print('n_cards: ', my_draw_pile.n_cards)
    my_card = my_draw_pile.give()
    print(my_card.colored())
    print(my_draw_pile.colored())
    print('n_cards: ', my_draw_pile.n_cards)

    while my_draw_pile.n_cards >= 1:
        my_draw_pile.give()
    print('\nOnce many cards are drawn..')
    print('n_cards: ', my_draw_pile.n_cards)
    print('my_draw_pile.give(): ', my_draw_pile.give())

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
from typing import Iterable, Tuple
from hanabython.Modules.Colored import Colored
from hanabython.Modules.Color import Color
from hanabython.Modules.Colors import Colors
from hanabython.Modules.ConfigurationColorContents \
    import ConfigurationColorContents
from collections import OrderedDict


class ConfigurationDeck(Colored, OrderedDict):
    """
    The contents of the deck for a game of Hanabi.

    This is essentially an OrderedDict. To each :class:`Color` in the deck, it
    associates the contents of the color, an object of class
    :class:`ConfigurationColorContents`.

    The order of the colors is important: it will be used in many occasions
    (including for display).

    :param contents: the iterable used to construct the ordered
        dictionary. Typically it is an :class:`OrderedDict` or a list of pairs
        (`color`, `contents`).
    :param name: the name of the configuration. Can be None (default value).
        Should not be capitalized (e.g. "my favorite configuration" and not
        "My favorite configuration").

    >>> cfg = ConfigurationDeck.NORMAL
    >>> print(cfg.name)
    normal
    >>> print(cfg)
    normal
    >>> cfg = ConfigurationDeck(
    ...     contents=[
    ...         (Colors.BLUE, ConfigurationColorContents.NORMAL),
    ...         (Colors.RED, ConfigurationColorContents([3, 2, 1]))
    ...     ]
    ... )
    >>> print(cfg.name)
    None
    >>> print(cfg)
    B normal, R [3, 2, 1]
    """
    def __init__(self,
                 contents: Iterable[Tuple[Color, ConfigurationColorContents]],
                 name: str = None):
        super(ConfigurationDeck, self).__init__(contents)
        self.name = name

    def colored(self) -> str:
        if self.name is None:
            return ', '.join([
                c.color_str('%s %s') % (c, v) for c, v in self.items()
            ])
        else:
            return self.name

    def copy(self) -> 'ConfigurationDeck':
        """
        Copy the deck configuration.

        :return: a copy of this deck configuration. You can modify the copy
            without modifying the original. However, it is not a deep copy,
            (most of time, it would not be useful).

        >>> cfg = ConfigurationDeck.NORMAL.copy()
        >>> cfg.name = None
        >>> del(cfg[Colors.WHITE], cfg[Colors.YELLOW])
        >>> print(cfg)
        B normal, G normal, R normal
        >>> print(ConfigurationDeck.NORMAL[Colors.WHITE])
        normal
        """
        return ConfigurationDeck(contents=self.items(), name=self.name)

    @staticmethod
    def normal_plus(
        contents: Iterable[Tuple[Color, ConfigurationColorContents]],
        name: str = None
    ) -> 'ConfigurationDeck':
        """
        Shortcut to define a deck configuration from the normal one.

        :param contents: the additional contents (typically multicolor, etc.)
        :param name: the name of the configuration.

        :return: the new configuration.

        >>> cfg = ConfigurationDeck.normal_plus(contents=[
        ...     (Colors.SIXTH, ConfigurationColorContents.NORMAL),
        ...     (Colors.MULTICOLOR, ConfigurationColorContents.SHORT)
        ... ])
        >>> print(cfg)
        B normal, G normal, R normal, W normal, Y normal, P normal, M short
        """
        result = ConfigurationDeck.NORMAL.copy()
        result.update(contents)
        result.name = name
        return result

    #: Normal deck (5 colors of 10 cards).
    NORMAL = None
    #: Deck with long sixth color (6 colors of 10 cards).
    W_SIXTH = None
    #: Deck with short sixth color (5 colors of 10 cards + 1 color of 5 cards).
    W_SIXTH_SHORT = None
    #: Deck with long multicolor (5 colors of 10 cards + 1 multi of 10 cards).
    W_MULTICOLOR = None
    #: Deck with short multicolor (5 colors of 10 cards + 1 multi of 5 cards).
    W_MULTICOLOR_SHORT = None
    #: Deck with 8 colors (6 colors + multicolor + colorless, all of 10 cards).
    EIGHT_COLORS = None


ConfigurationDeck.NORMAL = ConfigurationDeck(
    contents=[
        (c, ConfigurationColorContents.NORMAL)
        for c in [Colors.BLUE, Colors.GREEN, Colors.RED,
                  Colors.WHITE, Colors.YELLOW]
    ], name='normal'
)
ConfigurationDeck.W_SIXTH = ConfigurationDeck.normal_plus(
    contents=[(Colors.SIXTH, ConfigurationColorContents.NORMAL)],
    name='with normal sixth color (10 cards)'
)
ConfigurationDeck.W_SIXTH_SHORT = ConfigurationDeck.normal_plus(
    contents=[(Colors.SIXTH, ConfigurationColorContents.SHORT)],
    name='with short sixth color (5 cards)'
)
ConfigurationDeck.W_MULTICOLOR = ConfigurationDeck.normal_plus(
    contents=[(Colors.MULTICOLOR, ConfigurationColorContents.NORMAL)],
    name='with normal multicolor (10 cards)'
)
ConfigurationDeck.W_MULTICOLOR_SHORT = ConfigurationDeck.normal_plus(
    contents=[(Colors.MULTICOLOR, ConfigurationColorContents.SHORT)],
    name='with short multicolor (5 cards)'
)
ConfigurationDeck.EIGHT_COLORS = ConfigurationDeck.normal_plus(
    contents=[
        (Colors.SIXTH, ConfigurationColorContents.NORMAL),
        (Colors.MULTICOLOR, ConfigurationColorContents.NORMAL),
        (Colors.COLORLESS, ConfigurationColorContents.NORMAL)
    ],
    name='with sixth color, multicolor and colorless (10 cards each)'
)


if __name__ == '__main__':
    print('A deck configuration with a name:')
    my_cfg = ConfigurationDeck.W_MULTICOLOR
    my_cfg.test_str()

    print('\nA deck configuration with no name:')
    my_cfg = ConfigurationDeck(
        contents=[
            (Colors.BLUE, ConfigurationColorContents.NORMAL),
            (Colors.RED, ConfigurationColorContents([3, 2, 1]))
        ]
    )
    my_cfg.test_str()

    print('\nChange a part of the configuration:')
    my_cfg[Colors.BLUE] = ConfigurationColorContents([1, 1])
    print(my_cfg.colored())

    print('\nTest the copy method:')
    my_cfg = ConfigurationDeck.NORMAL.copy()
    my_cfg.name = None
    del(my_cfg[Colors.BLUE])
    print('New configuration:', my_cfg.colored())
    print('Blue in the old configuration:',
          ConfigurationDeck.NORMAL[Colors.BLUE])

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""


class StringAnsi:
    """
    An ANSI escape code that modifies the printing aspect.
    """

    #: This escape code is special: it is used to return to default aspect.
    RESET = "\033[0;0m"
    #:
    STYLE_BOLD = "\033[1m"
    #:
    STYLE_UNDERLINE = "\033[4m"
    #:
    STYLE_REVERSE_VIDEO = "\033[7m"

    WHITE_NOT_BRIGHT = "\033[30m"
    WHITE_BRIGHT = "\033[90m"
    #: This should be white on black background, and vice-versa.
    WHITE = WHITE_NOT_BRIGHT

    RED_NOT_BRIGHT = "\033[31m"
    RED_BRIGHT = "\033[91m"
    #:
    RED = RED_NOT_BRIGHT

    GREEN_NOT_BRIGHT = "\033[32m"
    GREEN_BRIGHT = "\033[92m"
    #:
    GREEN = GREEN_NOT_BRIGHT

    YELLOW_NOT_BRIGHT = "\033[33m"
    YELLOW_BRIGHT = "\033[93m"
    #:
    YELLOW = YELLOW_BRIGHT
    #:
    BROWN = YELLOW_NOT_BRIGHT

    BLUE_NOT_BRIGHT = "\033[34m"
    BLUE_BRIGHT = "\033[94m"
    #:
    BLUE = BLUE_BRIGHT

    MAGENTA_NOT_BRIGHT = "\033[35m"
    MAGENTA_BRIGHT = "\033[95m"
    #:
    MAGENTA = MAGENTA_NOT_BRIGHT

    CYAN_NOT_BRIGHT = "\033[36m"
    CYAN_BRIGHT = "\033[96m"
    #:
    CYAN = CYAN_BRIGHT


if __name__ == '__main__':
    # N.B.: Available colors for multi are brown, magenta and cyan
    for k in StringAnsi.__dict__.keys():
        if not k.startswith('_'):
            print(StringAnsi.__dict__[k] + k + StringAnsi.RESET)



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
from hanabython.Modules.StringAnsi import StringAnsi
from hanabython.Modules.Color import Color
from hanabython.Modules.ColorMulticolor import ColorMulticolor
from hanabython.Modules.ColorColorless import ColorColorless


class Colors:
    """
    Standard colors in Hanabi.
    """

    #:
    BLUE = Color(name='Blue', symbol='B', print_color=StringAnsi.BLUE)
    #:
    GREEN = Color(name='Green', symbol='G', print_color=StringAnsi.GREEN)
    #:
    RED = Color(name='Red', symbol='R', print_color=StringAnsi.RED)
    #:
    WHITE = Color(name='White', symbol='W', print_color=StringAnsi.WHITE)
    #:
    YELLOW = Color(name='Yellow', symbol='Y', print_color=StringAnsi.YELLOW)
    #: Use this for the sixth color. As of now, it is pink but the display
    #: color might change in future implementations.
    SIXTH = Color(name='Pink', symbol='P', print_color=StringAnsi.MAGENTA)
    #: Use this for multicolor cards. As of now, it is cyan but the display
    #: color might change in future implementations.
    MULTICOLOR = ColorMulticolor(
        name='Multicolor', symbol='M', print_color=(
            StringAnsi.CYAN + StringAnsi.STYLE_BOLD + StringAnsi.STYLE_UNDERLINE
        )
    )
    #: Use this for the colorless cards. As of now, it is brown but the display
    #: color might change in future implementations.
    COLORLESS = ColorColorless(
        name='Colorless', symbol='C', print_color=(
            StringAnsi.RED + StringAnsi.STYLE_BOLD + StringAnsi.STYLE_UNDERLINE
        )
    )

    @classmethod
    def from_symbol(cls, s: str) -> Color:
        """
        Find one of the standard colors from its symbol.

        :return: the corresponding color.

        >>> color = Colors.from_symbol('B')
        >>> print(color.name)
        Blue
        """
        for k in Colors.__dict__.keys():
            try:
                symbol = Colors.__dict__[k].symbol
            except AttributeError:
                continue
            if symbol == s:
                return Colors.__dict__[k]
        raise ValueError('Could not find color with symbol: ', s)


if __name__ == '__main__':
    Colors.BLUE.test_str()

    my_color = Colors.from_symbol('B')
    print('\n' + my_color.colored())
    try:
        my_color = Colors.from_symbol('Z')
        print(my_color.colored())
    except ValueError as e:
        print(e)

    print('\n' + Colors.BLUE.colored())
    print(Colors.GREEN.colored())
    print(Colors.RED.colored())
    print(Colors.WHITE.colored())
    print(Colors.YELLOW.colored())
    print(Colors.SIXTH.colored())
    print(Colors.MULTICOLOR.colored())
    print(Colors.COLORLESS.colored())

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
from hanabython.Modules.Clue import Clue
from hanabython.Modules.Colored import Colored
from hanabython.Modules.Configuration import Configuration
from hanabython.Modules.Color import Color
from hanabython.Modules.Colors import Colors
from hanabython.Modules.StringAnsi import StringAnsi


class CardPublic(Colored):
    """
    The "public" part of a card.

    An object of this class represents what is known by all players, including
    the owner of the card.

    :param cfg: the configuration of the game.

    :var np.array can_be_c: a coefficient is True iff the card can be of the
        corresponding color.
    :var np.array can_be_v: a coefficient is True iff the card can be of the
        corresponding value.
    :var np.array yes_clued_c: a coefficient is True iff the card was
        explicitly clued as the corresponding color *and* it can be of
        this color (this precision is important for multicolor).
    :var np.array yes_clued_v: a coefficient is True iff the card was
        explicitly clued as value v.

    >>> from hanabython.Modules.Configuration import Configuration
    >>> card = CardPublic(Configuration.EIGHT_COLORS)
    >>> print(card)
    BGRWYPMC 12345
    """
    def __init__(self, cfg: Configuration):
        self.cfg = cfg
        self.can_be_c = np.ones(cfg.n_colors, dtype=bool)       # type: np.array
        self.can_be_v = np.ones(cfg.n_values, dtype=bool)       # type: np.array
        self.yes_clued_c = np.zeros(cfg.n_colors, dtype=bool)   # type: np.array
        self.yes_clued_v = np.zeros(cfg.n_values, dtype=bool)   # type: np.array

    def colored(self) -> str:
        s_c = ''
        w_c = 0
        for i, c in enumerate(self.cfg.colors):
            if self.yes_clued_c[i]:
                s_c += StringAnsi.STYLE_REVERSE_VIDEO + c.color_str(c.symbol)
                w_c += 1
            elif self.can_be_c[i]:
                s_c += c.color_str(c.symbol)
                w_c += 1
        s_v = ''
        w_v = 0
        for i, v in enumerate(self.cfg.values):
            if self.yes_clued_v[i]:
                s_v += StringAnsi.STYLE_REVERSE_VIDEO + str(
                    v) + StringAnsi.RESET
                w_v += 1
            elif self.can_be_v[i]:
                s_v += str(v)
                w_v += 1
        if w_c == 1 and w_v == 1:
            s = s_c + s_v
            w = 2
        else:
            s = s_c + ' ' + s_v
            w = w_c + w_v + 1
        # Note: s.center(...) would not work because of the ANSI escape codes.
        display_width = self.cfg.n_colors + self.cfg.n_values + 1
        left = (display_width - w) // 2
        right = display_width - w - left
        return ' ' * left + s + ' ' * right

    def colored_old(self) -> str:
        s = ''
        for i, c in enumerate(self.cfg.colors):
            if self.yes_clued_c[i]:
                s += StringAnsi.STYLE_REVERSE_VIDEO + c.color_str(c.symbol)
            elif self.can_be_c[i]:
                s += c.color_str(c.symbol)
            else:
                s += ' '
        s += ' '
        for i, v in enumerate(self.cfg.values):
            if self.yes_clued_v[i]:
                s += StringAnsi.STYLE_REVERSE_VIDEO + str(v) + StringAnsi.RESET
            elif self.can_be_v[i]:
                s += str(v)
            else:
                s += ' '
        return s

    def _match_c(self, x: Color, b: bool) -> None:
        """
        React to a clue by color.

        Updates the internal variables of the card.

        :param x: the color of the clue.
        :param b: whether the card matched the clue or not.
        """
        for i, c in enumerate(self.cfg.colors):
            if c.match(x) != b:
                self.can_be_c[i] = False
                self.yes_clued_c[i] = False  # important for multicolor
            if b and c.match(x) and self.can_be_c[i]:
                self.yes_clued_c[i] = True

    def _match_v(self, x: int, b: bool) -> None:
        """
        React to a clue by value.

        Updates the internal variables of the card.

        :param x: the value of the clue.
        :param b: whether the card matched the clue or not.
        """
        i = self.cfg.i_from_v(x)
        if b:
            self.yes_clued_v[i] = True
            self.can_be_v[:] = False
            self.can_be_v[i] = True
        else:
            self.can_be_v[i] = False

    def match(self, clue: Clue, b: bool) -> None:
        """
        React to a clue.

        Updates the internal variables of the card.

        :param clue: the clue.
        :param b: whether the card matches or not.

        >>> from hanabython.Modules.Configuration import Configuration
        >>> cfg = Configuration.EIGHT_COLORS
        >>> card = CardPublic(cfg)
        >>> print(card)
        BGRWYPMC 12345
        >>> card.match(clue=Clue(Colors.RED), b=False)
        >>> print(card)  #doctest: +NORMALIZE_WHITESPACE
        BGWYPC 12345
        >>> card.match(clue=Clue(Colors.BLUE), b=True)
        >>> print(card)  #doctest: +NORMALIZE_WHITESPACE
        B 12345

        Let us try with the clues in the opposite order:

        >>> from hanabython.Modules.Configuration import Configuration
        >>> card = CardPublic(Configuration.EIGHT_COLORS)
        >>> print(card)
        BGRWYPMC 12345
        >>> card.match(clue=Clue(Colors.BLUE), b=True)
        >>> print(card)  #doctest: +NORMALIZE_WHITESPACE
        BM 12345
        >>> card.match(clue=Clue(Colors.RED), b=False)
        >>> print(card)  #doctest: +NORMALIZE_WHITESPACE
        B 12345

        Now with clues by value:

        >>> from hanabython.Modules.Configuration import Configuration
        >>> card = CardPublic(Configuration.EIGHT_COLORS)
        >>> print(card)
        BGRWYPMC 12345
        >>> card.match(clue=Clue(3), b=False)
        >>> print(card)  #doctest: +NORMALIZE_WHITESPACE
        BGRWYPMC 1245
        >>> card.match(clue=Clue(5), b=True)
        >>> print(card)  #doctest: +NORMALIZE_WHITESPACE
        BGRWYPMC 5
        """
        if clue.category == Clue.VALUE:
            self._match_v(clue.x, b)
        else:
            self._match_c(clue.x, b)


if __name__ == '__main__':
    my_card = CardPublic(Configuration.EIGHT_COLORS)
    my_card.test_str()

    print('\nIt is not red, then it is blue:')
    print(my_card.colored())
    my_card.match(clue=Clue(Colors.RED), b=False)
    print(my_card.colored())
    my_card.match(clue=Clue(Colors.BLUE), b=True)
    print(my_card.colored())

    print('\nIt is blue, then it is not red:')
    my_card = CardPublic(Configuration.EIGHT_COLORS)
    print(my_card.colored())
    my_card.match(clue=Clue(Colors.BLUE), b=True)
    print(my_card.colored())
    my_card.match(clue=Clue(Colors.RED), b=False)
    print(my_card.colored())

    print('\nIt is not 3, then it is 5:')
    my_card = CardPublic(Configuration.EIGHT_COLORS)
    print(my_card.colored())
    my_card.match(clue=Clue(3), b=False)
    print(my_card.colored())
    my_card.match(clue=Clue(5), b=True)
    print(my_card.colored())

    print('\nIt is 5, then it is not 3:')
    my_card = CardPublic(Configuration.EIGHT_COLORS)
    print(my_card.colored())
    my_card.match(clue=Clue(5), b=True)
    print(my_card.colored())
    my_card.match(clue=Clue(3), b=False)
    print(my_card.colored())

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
from hanabython.Modules.Colored import Colored
from hanabython.Modules.Configuration import Configuration


class DrawPilePublic(Colored):
    """
    The public part of a draw pile.

    An object of this class represents what is known by all players. In the
    normal version of the game and all official variants, it is only the number
    of cards left.

    :param cfg: the configuration of the game.

    >>> from hanabython.Modules.Configuration import Configuration
    >>> draw_pile = DrawPilePublic(cfg=Configuration.STANDARD)
    >>> print(draw_pile)
    50 cards left
    """
    def __init__(self, cfg: Configuration):
        self.cfg = cfg
        self.n_cards = cfg.n_cards

    def colored(self) -> str:
        if self.n_cards == 0:
            return 'No card left'
        if self.n_cards == 1:
            return '1 card left'
        return str(self.n_cards) + ' cards left'

    def give(self) -> None:
        """
        Give the card from the top of pile.

        Updates the internal variables of the pile.

        >>> from hanabython.Modules.Configuration import Configuration
        >>> draw_pile = DrawPilePublic(cfg=Configuration.STANDARD)
        >>> print(draw_pile)
        50 cards left
        >>> while draw_pile.n_cards >= 2:
        ...     draw_pile.give()
        >>> print(draw_pile)
        1 card left
        >>> draw_pile.give()
        >>> print(draw_pile)
        No card left
        """
        if self.n_cards > 0:
            self.n_cards -= 1


if __name__ == '__main__':
    my_draw_pile = DrawPilePublic(Configuration.STANDARD)
    my_draw_pile.test_str()

    print('\nDraw a card: ')
    my_draw_pile.give()
    print(my_draw_pile)
    while my_draw_pile.n_cards >= 2:
        my_draw_pile.give()
    print('\nOnce many cards are drawn..')
    print(my_draw_pile)
    print('Draw the last card...')
    my_draw_pile.give()
    print(my_draw_pile)
    print('Try to draw another card...')
    my_draw_pile.give()
    print(my_draw_pile)

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
from hanabython.Modules.StringUtils import uncolor


class Colored:
    r"""
    An object with a colored string representation.

    >>> from hanabython.Modules.StringAnsi import StringAnsi
    >>> class MyClass(Colored):
    ...     def colored(self):
    ...         return StringAnsi.RED + 'some text' + StringAnsi.RESET
    >>> my_object = MyClass()
    >>> my_object.colored()
    '\x1b[31msome text\x1b[0;0m'
    >>> str(my_object)
    'some text'
    >>> repr(my_object)
    '<MyClass: some text>'
    """

    def __repr__(self) -> str:
        return '<%s: %s>' % (self.__class__.__name__, self)

    def __str__(self) -> str:
        return uncolor(self.colored())

    def colored(self) -> str:
        """
        Colored version of :meth:`__str__`.

        In the subclasses, the principle is to override only this method.
        :meth:`__str__` is automatically defined as the uncolored version of
        the same string, and :meth:`__repr__` as the same with the class name
        added.

        Of course, it is also possible to override :meth:`__str__` and/or
        :meth:`__repr__` if a different behavior is desired.

        :return: a string representing the object, possibly with ANSI escape
            codes to add colors and style.
        :rtype: str
        """
        pass

    def test_str(self) -> None:
        """
        Test the string representations of the object.

        Print the results of :meth:`__repr__`, :meth:`__str__` and
        :meth:`colored`.
        """
        print('repr:')
        print(repr(self))
        print('\nstr:')
        print(self)
        print('\ncolored:')
        print(self.colored())


if __name__ == '__main__':
    from hanabython.Modules.StringAnsi import StringAnsi

    class TestClass(Colored):
        def colored(self):
            return StringAnsi.RED + 'some text' + StringAnsi.RESET
    test_object = TestClass()
    test_object.test_str()

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-

"""Top-level package for Hanabython."""

__author__ = """François Durand"""
__email__ = 'fradurand@gmail.com'
__version__ = '0.1.8'

from .Modules.Action import Action
from .Modules.ActionClue import ActionClue
from .Modules.ActionThrow import ActionThrow
from .Modules.ActionForfeit import ActionForfeit
from .Modules.ActionPlayCard import ActionPlayCard
from .Modules.Board import Board
from .Modules.Card import Card
from .Modules.CardPublic import CardPublic
from .Modules.Clue import Clue
from .Modules.Color import Color
from .Modules.ColorMulticolor import ColorMulticolor
from .Modules.ColorColorless import ColorColorless
from .Modules.Colors import Colors
from .Modules.Colored import Colored
from .Modules.Configuration import Configuration
from .Modules.ConfigurationColorContents import ConfigurationColorContents
from .Modules.ConfigurationDeck import ConfigurationDeck
from .Modules.ConfigurationEmptyClueRule import ConfigurationEmptyClueRule
from .Modules.ConfigurationEndRule import ConfigurationEndRule
from .Modules.ConfigurationHandSize import ConfigurationHandSize
from .Modules.DiscardPile import DiscardPile
from .Modules.DrawPile import DrawPile
from .Modules.DrawPilePublic import DrawPilePublic
from .Modules.Game import Game
from .Modules.Hand import Hand
from .Modules.HandPublic import HandPublic
from .Modules.Player import Player
from .Modules.PlayerBase import PlayerBase
from .Modules.PlayerHumanText import PlayerHumanText
from .Modules.PlayerPuppet import PlayerPuppet
from .Modules.StringAnsi import StringAnsi
from .Modules.StringUtils import uncolor, title, str_from_iterable



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
from hanabython.Modules.Action import Action


class ActionThrow(Action):
    """
    An action of a player: throw (discard willingly).

    :param k: position of the card in the hand (between 0 and `#cards - 1`).
        Be careful: as of now, :attr:`__str__` expresses the position in
        "user-friendly" format, i.e. between 1 and `#cards` (this behavior might
        change in the future).

    >>> action = ActionThrow(k=2)
    >>> print(action)
    Discard card in position 3
    """
    def __init__(self, k: int):
        super().__init__(Action.THROW)
        self.k = k

    def colored(self) -> str:
        # Reminded: in end-user interfaces (including ``colored`` methods),
        # we say 'discard'.
        return 'Discard card in position %s' % (self.k + 1)


if __name__ == '__main__':
    my_action = ActionThrow(k=2)
    my_action.test_str()

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
import re
from typing import Iterable


def uncolor(s: str) -> str:
    """
    Remove ANSI escape codes from the string.

    :param string s: a string.

    :return: the same string without its ANSI escape codes.

    >>> from hanabython.Modules.StringAnsi import StringAnsi
    >>> s = (StringAnsi.RED + "Hanabi" + StringAnsi.RESET + ', a game by '
    ...      + StringAnsi.BLUE + 'Antoine Bauza' + StringAnsi.RESET)
    >>> uncolor(s)
    'Hanabi, a game by Antoine Bauza'
    """
    return re.sub(r'\033.\d*(;\d*)?m', "", s)


def title(s: str, width: int) -> str:
    """
    Format a string as a title.

    :param s: the string
    :param width: the total width of the final layout (in number of
        characters).

    :return: the string formatted as a title.

    >>> title(s='Title', width=20)
    '****** Title *******'
    >>> title(s='A not-too-long title', width=20)
    'A not-too-long title'
    >>> title(s='A title that is really too long', width=20)
    'A title that is r...'
    """
    if len(s) > width:
        return s[:width - 3] + '...'
    if len(s) > width - 2:
        return s.ljust(width)
    left = (width - len(s)) // 2
    right = width - len(s) - left
    return '*' * (left - 1) + ' ' + s + ' ' + '*' * (right - 1)


def str_from_iterable(l: Iterable) -> str:
    """
    Convert an iterable to a simple string.

    There are two differences with the standard implementation of str:

    #. No brackets.
    #. For each ``item`` of the iterable, ``str_from_iterable`` uses
       ``str(item)``, whereas ``str`` uses ``repr(item)``.

    :param l: an iterable.
    :return: a simple string.

    >>> print(str_from_iterable(['a', 'b', 'c']))
    a b c
    >>> print(['a', 'b', 'c'])
    ['a', 'b', 'c']
    """
    return ' '.join([str(x) for x in l])


if __name__ == "__main__":
    my_s = "\033[0;31mHanabi\033[0;0m by \033[0;94mAntoine Bauza\033[0;0m"
    print(my_s)
    print(uncolor(my_s))

    print()
    print(title(s='Title', width=20))
    print(title(s='123456789012345678', width=20))
    print(title(s='1234567890123456789', width=20))
    print(title(s='12345678901234567890', width=20))
    print(title(s='123456789012345678901', width=20))
    print(title(s='A title that is really too long', width=20))

    print()
    print(str_from_iterable(['a', 'b', 'c']))
    print(['a', 'b', 'c'])

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
from hanabython.Modules.Colored import Colored


class ConfigurationEndRule(Colored):
    """
    A rule for the end of game in Hanabi.

    This class does not implement the rules themselves: they are hardcoded in
    the class :class:`Game`.

    :param i: a unique identifier of the rule.
    :param name: the name of the configuration.
        Should not be capitalized (e.g. "my favorite configuration" and not
        "My favorite configuration"), except if it is seen as a title
        (e.g. "Crowning Piece").

    >>> cfg = ConfigurationEndRule.NORMAL
    >>> print(cfg)
    normal
    >>> print(cfg==ConfigurationEndRule.NORMAL)
    True
    >>> print(cfg==ConfigurationEndRule.CROWNING_PIECE)
    False
    """

    def __init__(self, i: int, name: str):
        self.i = i
        self.name = name

    def colored(self) -> str:
        return self.name

    def __eq__(self, other: 'ConfigurationEndRule') -> bool:
        return self.i == other.i

    #: Default rule for the end of game. When a player draws the last card,
    #: all players play one last time (her included).
    NORMAL = None
    #: "Crowning piece" variant for the end of game. The game stops when a
    #: player starts her turn with no card in hand.
    CROWNING_PIECE = None


ConfigurationEndRule.NORMAL = ConfigurationEndRule(0, 'normal')
ConfigurationEndRule.CROWNING_PIECE = ConfigurationEndRule(1, 'Crowning Piece')


if __name__ == '__main__':
    cfg = ConfigurationEndRule.NORMAL
    cfg.test_str()

    print('\nIs it normal?', cfg == ConfigurationEndRule.NORMAL)
    print('Is it Crowning Piece?', cfg == ConfigurationEndRule.CROWNING_PIECE)

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
from hanabython.Modules.Colored import Colored


class ConfigurationEmptyClueRule(Colored):
    """
    A rule for "empty clues" in Hanabi.

    An empty clue is a clue that corresponds to 0 cards in the hand of the
    concerned partner.

    This class does not implement the rules themselves: they are hardcoded in
    the class :class:`Game`.

    :param i: a unique identifier of the rule.
    :param name: the name of the configuration.
        Should not be capitalized (e.g. "my favorite configuration" and not
        "My favorite configuration").

    >>> cfg = ConfigurationEmptyClueRule.FORBIDDEN
    >>> print(cfg)
    empty clues are forbidden
    >>> print(cfg==ConfigurationEmptyClueRule.FORBIDDEN)
    True
    >>> print(cfg==ConfigurationEmptyClueRule.ALLOWED)
    False
    """

    def __init__(self, i: int, name: str):
        self.i = i
        self.name = name

    def colored(self) -> str:
        return self.name

    def __eq__(self, other: 'ConfigurationEmptyClueRule') -> bool:
        return self.i == other.i

    #:
    FORBIDDEN = None
    #:
    ALLOWED = None


ConfigurationEmptyClueRule.FORBIDDEN = ConfigurationEmptyClueRule(
    0, 'empty clues are forbidden')
ConfigurationEmptyClueRule.ALLOWED = ConfigurationEmptyClueRule(
    1, 'empty clues are allowed')


if __name__ == '__main__':
    cfg = ConfigurationEmptyClueRule.ALLOWED
    cfg.test_str()

    print('\nIs it allowed?', cfg == ConfigurationEmptyClueRule.ALLOWED)
    print('Is it forbidden?', cfg == ConfigurationEmptyClueRule.FORBIDDEN)

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
from typing import Iterable, Union, List
from hanabython.Modules.Clue import Clue
from hanabython.Modules.Colored import Colored
from hanabython.Modules.Card import Card
from hanabython.Modules.Colors import Colors


class Hand(Colored, list):
    """
    The hand of a player.

    We use the same convention as in Board Game Arena: newest cards are on the
    left (i.e. at the beginning of the list) and oldest cards are on the right
    (i.e. at the end of the list).

    Basically, a Hand is a list of Card objects. It can be constructed as such,
    or using a list of strings which will be automatically converted to cards.

    :param source: an iterable used to construct the hand. N.B.: this parameter
        is mostly used for examples and tests. In contrast, at the beginning of
        a game, the hand should be initialized with no cards, because cards will
        be given one by one to the players during the initial dealing of hands.

    >>> hand = Hand([Card('Y3'), Card('M1'), Card('B2'), Card('R4')])
    >>> print(hand)
    Y3 M1 B2 R4
    >>> hand = Hand(['Y3', 'M1', 'B2', 'R4'])
    >>> print(hand)
    Y3 M1 B2 R4
    """
    def __init__(self, source: Iterable[Union[Card, str]] = None):
        super().__init__()
        if source is not None:
            for item in source:
                if type(item) == str:
                    self.append(Card(item))
                else:
                    self.append(item)

    def colored(self) -> str:
        return ' '.join(card.colored() for card in self)

    def receive(self, card: Card) -> None:
        """
        Receive a card.

        :param card: the card received.

        The card is added on the left, i.e. at the beginning of the list.

        >>> hand = Hand(['Y3', 'M1', 'B2', 'R4'])
        >>> hand.receive(Card('G2'))
        >>> print(hand)
        G2 Y3 M1 B2 R4
        """
        self.insert(0, card)

    def give(self, k: int) -> Card:
        """
        Give a card.

        :param k: the position of the card in the hand (0 = newest).

        :return: the card given.

        >>> hand = Hand(['Y3', 'B1', 'M1', 'B2', 'R4'])
        >>> card = hand.give(1)
        >>> print(card)
        B1
        >>> print(hand)
        Y3 M1 B2 R4
        """
        return self.pop(k)

    def match(self, clue: Clue) -> List[bool]:
        """
        React to a clue.

        :param clue: the clue.

        :return: a list of booleans. The `i`-th coefficient is `True`
            iff the `i`-th card of the hand matches the clue given.

        >>> hand = Hand(['G2', 'Y3', 'M1', 'B2', 'R4'])
        >>> hand.match(Clue(Colors.RED))
        [False, False, True, False, True]
        >>> hand.match(Clue(2))
        [True, False, False, True, False]
        """
        return [card.match(clue) for card in self]


if __name__ == '__main__':
    my_hand = Hand(['Y3', 'B1', 'M1', 'B2', 'R4'])
    my_hand.test_str()

    my_card = my_hand.give(1)
    print('\nCard given: ', my_card.colored())
    print(my_hand.colored())

    my_card = Card('G2')
    my_hand.receive(my_card)
    print('\nCard received: ', my_card.colored())
    print(my_hand.colored())

    print('\nMatch red clue:')
    print(my_hand.match(Clue(Colors.RED)))

    print('\nMatch clue 2:')
    print(my_hand.match(Clue(2)))
    # print(hand.bool_list_from_clue(Action(
    #     category=Action.INFORM, clue_type=Action.VALUE, clue=2
    # )))

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Copyright François Durand
fradurand@gmail.com

This file is part of Hanabython.

    Hanabython is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hanabython is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hanabython.  If not, see <http://www.gnu.org/licenses/>.
"""
from hanabython.Modules.Action import Action
from hanabython.Modules.Clue import Clue


class ActionClue(Action):
    """
    An action of a player: give a clue.

    :param i: the relative position of the concerned player
        (i.e. 1 for next player, 2 for second next player, etc.).
    :param clue: the clue.

    >>> from hanabython.Modules.Colors import Colors
    >>> action = ActionClue(i=1, clue=Clue(2))
    >>> print(action)
    Clue 2 to player in relative position 1
    >>> action = ActionClue(i=2, clue=Clue(Colors.BLUE))
    >>> print(action)
    Clue B to player in relative position 2
    """

    def __init__(self, i: int, clue: Clue):
        super().__init__(Action.CLUE)
        self.i = i
        self.clue = clue

    def colored(self) -> str:
        return 'Clue %s to player in relative position %s' % (
            self.clue.colored(), self.i)


if __name__ == '__main__':
    my_action = ActionClue(i=1, clue=Clue(3))
    my_action.test_str()

    print()
    from hanabython.Modules.Colors import Colors
    my_action = ActionClue(i=1, clue=Clue(Colors.BLUE))
    my_action.test_str()

    import doctest
    doctest.testmod()
